from typing import Any, Dict, List
from urllib.parse import quote
import os
import requests
import json

# /*******************************************************************************
#  * Copyright (c) 2019 - 2022 Cambridge Semantics Incorporated.
#  * All rights reserved.
#  * 
#  * Contributors:
#  *     Cambridge Semantics Incorporated
#  *******************************************************************************/

class AnzoRequest:
    """Encodes a request against Anzo and supports executing the request.

    Instances of AnzoRequest are typically constructed with an
    AnzoRequestBuilder. When a request is executed, a RunTimeError may be
    raised depending on the status code of the HTTP response.

    Attributes:
        url: The url for connecting to the anzo server
        username: Username of the user with which to interact with Anzo.
        password: The password of the user
        headers: The HTTP headers included in the request
        data: The body of the HTTP request
        timeout_seconds: Timeout for the request in seconds
    """

    COOKIE_KEY = "Cookie"

    def __init__(self, url: str,
                 username: str,
                 password: str,
                 auth_token: str,
                 headers: Dict[str, Any],
                 data: Any,
                 timeout_seconds: float) -> None:
        self.url = url
        self.username = username
        self.password = password
        self.auth_token = auth_token
        self.headers = headers
        self.data = data
        self.timeout_seconds = timeout_seconds

    def execute_request(self) -> str:
        """Executes the anzo request and returns the body of the response.

        Returns:
            The body of the HTTP response.

        Raises:
            RunTimeError: If anything goes wrong in executing the request
                or if a non-success HTTP response is received.
        """

        try:
            if self.username and self.password:
                authorization = (self.username, self.password)
            else:
                authorization = ("", "")

            if self.auth_token:
                self.headers[self.COOKIE_KEY] = self.auth_token

            response = requests.post(
                self.url,
                headers=self.headers,
                data=self.data,
                auth=authorization,
                verify=False,
                timeout=self.timeout_seconds
            )

        except requests.exceptions.ReadTimeout:
            msg = (
                f"Request timed out with the timeout set to "
                f"{self.timeout_seconds} seconds"
            )
            raise TimeoutError(msg) from None
        except Exception as e:
            raise RuntimeError from e

        if response.status_code > 299:
            error = (
                f"Response from {self.url} returned with "
                f"non-success status code {response.status_code} "
                f"and response text: {response.text}"
            )

            raise RuntimeError(error)
        return response.text

    def __str__(self) -> str:
        return (
            f"<AnzoRequest url='{self.url}, "
            f"username='{self.username}', "
            f"password='{self.password}'>"
        )


class AnzoRequestBuilder:
    """Builds an AnzoRequest for executing a request against Anzo.

    Builds an AnzoRequest step-by-step. Once a request is fully specified,
    build() returns an AnzoRequest object which can be executed.

    Attributes:
        payload_dict: The body of the HTTP request. If payload_dict is defined,
            then payload_list must be empty
        payload_list: The body of the HTTP request. If payload_list is defined,
            then payload_dict must be empty
        base_url: The base url used to connect to the anzo server
        url_postfix: A postfix to the url that specifies the type of request
            (e.g. query, semantic service, wtc.)
        username: Username of the user with which to interact with Anzo.
        password: The password of the user
        headers: The HTTP headers included in the request
        timeout_seconds: Timeout for the request in seconds

    Usage:
        arb = AnzoRequestBuilder()
        arb.with_url(self.server, self.port)
        arb.with_auth(self.username, self.password, self.auth_token)
        arb.with_graphmart(graphmart)
        arb.with_query_string(query_string)
        anzo_request = arb.build()
        response = anzo_request.execute_request()
    """

    JSON_FORMAT = 'JSON'  # this is case sensitive
    TEXT_JSON_FORMAT = 'text/json'
    APP_JSON_FORMAT = 'application/json'

    PAYLOAD_QUERY_KEY = "query"
    PAYLOAD_FORMAT_KEY = "format"
    PAYLOAD_UPDATE_KEY = "update"
    DEFAULT_NAMED_GRAPH_KEY = "default-named-graph-uri"
    NAMED_GRAPH_KEY = "named-graph-uri"
    DEFAULT_GRAPH_KEY = "default-graph-uri"
    USING_GRAPH_KEY = "using-graph-uri"
    USING_NAMED_GRAPH_KEY = "using-named-graph-uri"
    URI_KEY = "uri"

    JOURNAL_QUERY_POSTFIX = "sparql"
    GRAPHMART_QUERY_POSTFIX = "sparql/graphmart"
    LDS_QUERY_POSTFIX = "sparql/lds"
    SEMANTIC_SERVICE_POSTFIX = "anzoclient/call"
    GET_POSTFIX = "anzoclient/get"
    ADD_POSTFIX = "anzoclient/add"
    REMOVE_POSTFIX = "anzoclient/remove"

    HEADER_URI_KEY = "uri"
    HEADER_FORMAT_KEY = "format"
    HEADER_CONTENT_TYPE_KEY = "Content-Type"
    SKIP_CACHE = "skipCache"

    def __init__(self) -> None:
        self.payload_dict: Dict[str, Any] = {}
        self.payload_list: List = []
        self.headers: Dict[str, Any] = {}
        self.base_url = ""
        self.url_postfix = ""
        self.username = ""
        self.password = ""
        self.auth_token = ""
        self.timeout_seconds = None

    def _raise_if_url_postfix_is_set(self) -> None:
        if self.url_postfix:
            raise RuntimeError("URL Postfix has already been set")

    def with_url(self, url: str, port: str) -> None:
        """Adds the url to the AnzoRequestBuilder.

        Raises:
            RuntimeError: If the url has already been set.
        """

        if self.base_url:
            raise RuntimeError("Base URL has already been set")

        self.base_url = f'https://{url}:{port}'

    def with_auth(self, username: str, password: str, auth_token: str) -> None:
        """Adds either a username and password or an auth_token
           to the AnzoRequestBuilder.

        Raises:
            RuntimeError: If the username, password, or
                          auth_token has already been set.
            ValueError: If it is not the case that either a username and
                        password have been specified or an auth_token
                        has been specified then an a ValueError is raised.
        """

        if self.username or self.password or self.auth_token:
            raise RuntimeError("Username or password has already been set")

        if username and password and auth_token:
            error = ("Only one of username and password"
                     " or auth_token may be specified")
            raise ValueError(error)

        if not (username or password or auth_token):
            error = "Some form of authentication must be used."
            raise ValueError(error)

        if (username and not password) or (password and not username):
            error = "Both a username and a password must be specified"
            raise ValueError(error)

        if username and password:
            self.username = username
            self.password = password
        else:
            self.auth_token = auth_token

    def with_timeout_seconds(self, timeout_seconds: float) -> None:
        self.timeout_seconds = timeout_seconds

    def with_query_string(self, query_string: str) -> None:
        """Adds the query to the AnzoRequestBuilder.
        """

        self.payload_dict[self.PAYLOAD_QUERY_KEY] = query_string
        self.payload_dict[self.PAYLOAD_FORMAT_KEY] = self.TEXT_JSON_FORMAT

    def with_query_file(self, query_file: str) -> None:
        """Adds the query to the AnzoRequestBuilder.
        """

        _, query_file_extension = os.path.splitext(query_file)
        if query_file_extension != '.rq':
            raise RuntimeError("Query Filepaths must end in '.rq'")

        with open(query_file, 'r') as query_file_handle:
            query_string = query_file_handle.read()
            self.payload_dict[self.PAYLOAD_QUERY_KEY] = query_string
        self.payload_dict[self.PAYLOAD_FORMAT_KEY] = self.TEXT_JSON_FORMAT

    def with_update_string(self, query_string: str) -> None:
        """Adds the query to the AnzoRequestBuilder.
        """

        self.payload_dict[self.PAYLOAD_UPDATE_KEY] = query_string
        self.payload_dict[self.PAYLOAD_FORMAT_KEY] = self.TEXT_JSON_FORMAT

    def with_update_file(self, query_file: str) -> None:
        """Adds the query to the AnzoRequestBuilder.
        """

        _, query_file_extension = os.path.splitext(query_file)
        if query_file_extension != '.rq':
            raise RuntimeError("Query Filepaths must end in '.rq'")

        with open(query_file, 'r') as query_file_handle:
            query_string = query_file_handle.read()
            self.payload_dict[self.PAYLOAD_UPDATE_KEY] = query_string
        self.payload_dict[self.PAYLOAD_FORMAT_KEY] = self.TEXT_JSON_FORMAT

    def with_graphmart(self, graphmart_uri: str,
                       named_graphs: List[str] = None) -> None:
        """Adds the graphmart postfix and uri to the AnzoRequestBuilder.
        """

        encoded_graphmart_uri = quote(graphmart_uri, safe='')
        self.url_postfix = (
            f"{self.GRAPHMART_QUERY_POSTFIX}/{encoded_graphmart_uri}"
        )

        if named_graphs:
            self.payload_dict[self.DEFAULT_GRAPH_KEY] = named_graphs
            self.payload_dict[self.NAMED_GRAPH_KEY] = named_graphs

    def with_query_journal(self, named_graphs: List[str] = None) -> None:
        """Adds the journal postfix to the AnzoRequestBuilder and sets up
        for a query against the journal.

        Raises:
            RuntimeError: If the postfix has already been set.
        """

        self._raise_if_url_postfix_is_set()
        self.url_postfix = self.JOURNAL_QUERY_POSTFIX

        # Don't serialize the named graphs to a single string - we want
        # multiple graphs to be included under the same header, which
        # requires a list. See
        # https://stackoverflow.com/questions/23384230/how-to-post-multiple-value-with-same-key-in-python-requests  # noqa
        # for more info.

        if named_graphs:
            self.payload_dict[self.DEFAULT_NAMED_GRAPH_KEY] = named_graphs
            self.payload_dict[self.NAMED_GRAPH_KEY] = named_graphs

    def with_update_journal(self, named_graphs: List[str] = None) -> None:
        """Adds the journal postfix to the AnzoRequestBuilder and sets up
        for an update query against the journal.

        Raises:
            RuntimeError: If the postfix has already been set.
        """

        self._raise_if_url_postfix_is_set()
        self.url_postfix = self.JOURNAL_QUERY_POSTFIX

        # Don't serialize the named graphs to a single string - we want
        # multiple graphs to be included under the same header, which
        # requires a list. See
        # https://stackoverflow.com/questions/23384230/how-to-post-multiple-value-with-same-key-in-python-requests  # noqa
        # for more info.

        if named_graphs:
            self.payload_dict[self.USING_GRAPH_KEY] = named_graphs
            self.payload_dict[self.USING_NAMED_GRAPH_KEY] = named_graphs

    def with_lds_cat_entry(self, lds_cat_entry: str,
                           named_graphs: List[str] = None) -> None:
        """Adds the lds postfix and uri to the AnzoRequestBuilder.

        Raises:
            RuntimeError: If the postfix has already been set.
        """

        self._raise_if_url_postfix_is_set()
        encoded_lds_cat_entry = quote(lds_cat_entry, safe='')
        self.url_postfix = f'{self.LDS_QUERY_POSTFIX}/{encoded_lds_cat_entry}'

        if named_graphs:
            self.payload_dict[self.DEFAULT_GRAPH_KEY] = named_graphs
            self.payload_dict[self.NAMED_GRAPH_KEY] = named_graphs

    def with_semantic_service(self, service_uri: str,
                              request: List) -> None:
        """Adds the semantic service postfix and payload to the AnzoRequestBuilder.

        Raises:
            RuntimeError: If the postfix has already been set.
        """

        self._raise_if_url_postfix_is_set()
        self.url_postfix = self.SEMANTIC_SERVICE_POSTFIX
        self.headers[self.HEADER_URI_KEY] = service_uri
        self.headers[self.HEADER_FORMAT_KEY] = self.JSON_FORMAT
        self.headers[self.HEADER_CONTENT_TYPE_KEY] = self.APP_JSON_FORMAT
        self.payload_list = request

    def with_get(self, named_graph_uri: str) -> None:
        """Adds the get postfix and named_graph_uri to the payload

        Raises:
            RuntimeError: If the postfix has already been set.
        """

        self._raise_if_url_postfix_is_set()
        self.url_postfix = self.GET_POSTFIX
        self.payload_dict[self.URI_KEY] = named_graph_uri

    def with_add(self, request: List) -> None:
        """Adds the add postfix and payload to the AnzoRequestBuilder.

        Raises:
            RuntimeError: If the postfix has already been set.
        """

        self._raise_if_url_postfix_is_set()
        self.url_postfix = self.ADD_POSTFIX
        self.headers[self.HEADER_FORMAT_KEY] = self.JSON_FORMAT
        self.headers[self.HEADER_CONTENT_TYPE_KEY] = self.APP_JSON_FORMAT
        self.payload_list = request

    def with_remove(self, request: List) -> None:
        """Adds the remove postfix and payload to the AnzoRequestBuilder.

        Raises:
            RuntimeError: If the postfix has already been set.
        """

        self._raise_if_url_postfix_is_set()
        self.url_postfix = self.REMOVE_POSTFIX
        self.headers[self.HEADER_FORMAT_KEY] = self.JSON_FORMAT
        self.headers[self.HEADER_CONTENT_TYPE_KEY] = self.APP_JSON_FORMAT
        self.payload_list = request

    def with_cache_skipped(self, skip_cache: bool):
        """Adds the option to skip AnzoGraph's query cache.

        Raises:
            RuntimeError: We currently don't support skipping the cache
                   so if a user tries to do it, an error is raised.
        """
        if skip_cache:
            self.payload_dict[self.SKIP_CACHE] = "true"

    def build(self) -> AnzoRequest:
        """
        Combine all of the relevant components to build an AnzoRequest.

        Returns:
            A fully developed AnzoRequest object

        Raises:
            ValueError: If any of the relevant components (base_url,
                url_postfix, username and password or authentication,
                and payload_dict OR
                payload_list) of the request are missing.
        """

        if not self.base_url:
            raise ValueError(
                "Can't build an AnzoRequest because "
                "self.base_url is missing"
            )

        if not self.url_postfix:
            raise ValueError(
                "Can't build an AnzoRequest because "
                "self.url_postfix is missing"
            )

        if not((self.username and self.password) or self.auth_token):
            raise ValueError(
                "Can't build AnzoRequest because authorization is missing"
            )

        if self.payload_dict and self.payload_list:
            raise ValueError(
                "Both self.payload_dict and self.payload_list"
                "cannot be populated"
            )

        if not self.timeout_seconds:
            raise ValueError("A timeout for the request must be specified")

        # Set the data of the request based on the payload.
        # Python-Requests is generous with the type that data can be,
        # so we have two cases:
        # (a) The payload is a dict in which case we pass data as a dict
        # (b) the payload is a list, in which case we pass the serialied
        #     list as a string

        data: Any = None
        if self.payload_dict:
            data = self.payload_dict
        elif self.payload_list:
            data = json.dumps(self.payload_list)

        full_url = f"{self.base_url}/{self.url_postfix}"

        return AnzoRequest(
            full_url, self.username, self.password, self.auth_token,
            self.headers, data, self.timeout_seconds
        )



# /*******************************************************************************
#  * Copyright (c) 2019 - 2022 Cambridge Semantics Incorporated.
#  * All rights reserved.
#  * 
#  * Contributors:
#  *     Cambridge Semantics Incorporated
#  *******************************************************************************/

import unittest
import time
import os

from pyanzo import (
    AnzoClient,
    QuadStore
)

from pyanzo.orchestration_wrapper import (
    OrchestrationWrapper,
    IngestionComponentStatus,
)

from .test_common import (
    GRAPHMART,
    SERVER,
    PORT,
    USERNAME,
    PASSWORD,
    log_while_running_tests,
)

# log_while_running_tests()

class TestOrchestrationWrapper(unittest.TestCase):
    im_uri = "http://cambridgesemantics.com/service/devIngestManager"

    def setUp(self) -> None:
        self.anzo_client = AnzoClient(
            server=SERVER, username=USERNAME, password=PASSWORD, port=PORT
        )

    @unittest.skip("Integration Test")
    def test_orchestration_wrapper(self):
        osm = OrchestrationWrapper(self.anzo_client, self.im_uri, [])

        osm.start()
        run_status = osm.block_until_complete(2000)

    def test_ingestion_component_status_construction1(self) -> None:

        ing_comp_trig = """
            @prefix ns1: <urn://etl/> .
            @prefix ns2: <http://openanzo.org/ontologies/2008/07/ASDL/IngestManager#> .

            ns1:4f558d33-dadd-4cdb-b4a4-7d94e069423f {
                   ns1:4f558d33-dadd-4cdb-b4a4-7d94e069423f a ns2:IngestionComponent ;
                       ns2:componentDescription "NULL" ;
                       ns2:componentLoadService <http://cambridgesemantics.com/Project/6bdebd2d-a585-480c-31d2-ff1c39eb3130/6bdebd2d-a585-480c-31d2-ff1c39eb3130/7696237e-1c7b-c488-6a6d-       9bdd6bdfb6b2/Job/919fb444-80e7-19b5-25a5-63b6903a2ba1> ;
                       ns2:componentName "Load sample" ;
                       ns2:componentStatus "STARTING" .
             }
             """  # noqa
        uri = "urn://etl/4f558d33-dadd-4cdb-b4a4-7d94e069423f"

        ing_comp_qs = QuadStore.from_trig_string(ing_comp_trig)
        ics = IngestionComponentStatus.from_quadstore(ing_comp_qs, uri)
        ics2 = IngestionComponentStatus(
            uri, "Load sample", "STARTING", False, "", "", "", ""
        )

        self.assertEqual(ics, ics2)

    def test_ingestion_component_status_construction2(self):

        ing_comp_trig = """
            @prefix ns1: <urn://etl/> .
            @prefix ns2: <http://openanzo.org/ontologies/2008/07/ASDL/IngestManager#> .
            @prefix xsd: <http://www.w3.org/2001/XMLSchema#> .

            ns1:d4f6d003-0545-4039-87ae-4da95fafca06 {
                ns1:d4f6d003-0545-4039-87ae-4da95fafca06 a ns2:IngestionComponent,
                        <http://openanzo.org/ontologies/2008/07/ASDL/IngestManager/EtlLoadService#StructuredIngestionComponent> ;
                    ns2:componentDescription "Job began at 2021-04-27T14:32:56.071Z and ended at 2021-04-27T14:33:02.039Z." ;
                    ns2:componentLoadService <http://cambridgesemantics.com/Project/6bdebd2d-a585-480c-31d2-ff1c39eb3130/6bdebd2d-a585-480c-31d2-ff1c39eb3130/7696237e-1c7b-c488-6a6d-9bdd6bdfb6b2/Job/919fb444-80e7-19b5-25a5-63b6903a2ba1> ;
                    #ns2:componentName "ETL ingestion component: Loadsample_sample" ;
                    ns2:componentName "" ;
                    ns2:componentStatus "success" ;
                    ns2:componentSucceeded true ;
                    ns2:contextGraph <http://cambridgesemantics.com/Project/6bdebd2d-a585-480c-31d2-ff1c39eb3130/6bdebd2d-a585-480c-31d2-ff1c39eb3130/7696237e-1c7b-c488-6a6d-9bdd6bdfb6b2/Job/919fb444-80e7-19b5-25a5-63b6903a2ba1/6> ;
                    ns2:index "1"^^xsd:int ;
                    ns2:ingestEndTime "2021-04-27T14:33:02.409000+00:00"^^xsd:dateTime ;
                    ns2:ingestStartTime "2021-04-27T14:32:49.129000+00:00"^^xsd:dateTime ;
                    ns2:newFldsUri <http://csi.com/FileBasedLinkedDataSet/faf590ddadf853e4d213b8ac823eb194> ;
                    ns2:percentComplete 1e+00 ;
                    ns2:previousFldsUri <http://csi.com/FileBasedLinkedDataSet/faf590ddadf853e4d213b8ac823eb194> .
            }

            """  # noqa
        uri = "urn://etl/4f558d33-dadd-4cdb-b4a4-7d94e069423f"

        ing_comp_qs = QuadStore.from_trig_string(ing_comp_trig)
        ics = IngestionComponentStatus.from_quadstore(ing_comp_qs, uri)
        ics2 = IngestionComponentStatus(
            uri, "", "success", True, "2021-04-27T14:32:49.129000+00:00",
            "2021-04-27T14:33:02.409000+00:00", "1", "1.0"
        )

        self.assertEqual(ics, ics2)



# /*******************************************************************************
#  * Copyright (c) 2019 - 2022 Cambridge Semantics Incorporated.
#  * All rights reserved.
#  * 
#  * Contributors:
#  *     Cambridge Semantics Incorporated
#  *******************************************************************************/

from typing import List, Set, Dict, Any
import rdflib

from .uris import NAMESPACE


def semantic_type_to_python_object(val: str, val_datatype: str) -> any:
    """
    Returns val as the python type specified by val_datatype. For example,
    if val_datatype is xsd:int, then val parsed as an int will be returned.

    Args:
        val: The value to parse
        val_datatype: The datatype of val, used to inform the parsing

    Returns: val parsed as a python object; parsed as indicated
        by the val_datatype
    """

    mapping = {
        "http://www.w3.org/2001/XMLSchema#int": int,
        "http://www.w3.org/2001/XMLSchema#integer": int,
        "http://www.w3.org/2001/XMLSchema#long": int,
        "http://www.w3.org/2001/XMLSchema#float": float,
        "http://www.w3.org/2001/XMLSchema#double": float,
        "http://www.w3.org/2001/XMLSchema#decimal": float,
    }

    if val_datatype in mapping:
        func = mapping[val_datatype]
        return func(val)
    return val


class Quad:
    NAMED_GRAPH_URI_KEY = "namedGraphUri"
    SUBJECT_KEY = "subject"
    VALUE_KEY = "value"
    OBJECT_KEY = "object"
    PREDICATE_KEY = "predicate"
    TYPE_KEY = "objectType"
    DATA_TYPE_KEY = "dataType"

    def __init__(self, sub, pred, obj, graph, sub_type="", obj_type="", obj_data_type=""):
        self.sub = sub
        self.pred = pred
        self.obj = obj
        self.graph = graph

        self.sub_type = sub_type if sub_type else "uri"

        # obj_type is typically either uri or literal
        self.obj_type = obj_type if obj_type else "uri"

        # obj_data_type is only non-empty if obj_type
        # is a literal, and it will be the type of literal,
        # like an xsd:int or xsd:float.
        self.obj_data_type = obj_data_type

    @staticmethod
    def from_anzo_json_dict(item: Dict[str, Any]):
        try:
            sub = item[Quad.SUBJECT_KEY][Quad.VALUE_KEY]
            sub_type = item[Quad.SUBJECT_KEY][Quad.TYPE_KEY]

            pred = item[Quad.PREDICATE_KEY]

            obj = item[Quad.OBJECT_KEY][Quad.VALUE_KEY]
            obj_type = item[Quad.OBJECT_KEY][Quad.TYPE_KEY]

            obj_data_type = item[Quad.OBJECT_KEY].get(Quad.DATA_TYPE_KEY, "")

        except (KeyError, TypeError, AttributeError) as e:
            raise RuntimeError(f"The JSON object {item} is not suited to"
                               " be transformed into a Quad. ") from e

        graph = item.get(Quad.NAMED_GRAPH_URI_KEY, "")
        return Quad(sub, pred, obj, graph, sub_type, obj_type, obj_data_type)

    def as_anzo_json_dict(self) -> dict:

        object_dict = dict()

        object_dict = {
            self.VALUE_KEY: self.obj,
            self.TYPE_KEY: self.obj_type
        }

        if self.obj_data_type:
            object_dict[self.DATA_TYPE_KEY] = self.obj_data_type

        return {
            self.NAMED_GRAPH_URI_KEY: self.graph,
            self.SUBJECT_KEY: {
                self.VALUE_KEY: self.sub,
                self.TYPE_KEY: self.sub_type
            },
            self.PREDICATE_KEY: self.pred,
            self.OBJECT_KEY: object_dict,
        }

    def as_record_dict(self):
        ret = {
            's': semantic_type_to_python_object(self.sub, self.sub_type),
            'p': self.pred,
            'o': semantic_type_to_python_object(self.obj, self.obj_data_type)
        }

        # Only add the graph if it exists
        if self.graph:
            ret['g'] = self.graph
        return ret

    def verbosify(self) -> str:
        """Returns a verbose string describing the quad. For a less
        verbose serialization, use str()
        """

        return (
            f"<Quad sub={self.sub}, sub_type={self.sub_type} "
            f"pred={self.pred}, "
            f"obj={self.obj}, obj_type={self.obj_type}, "
            f"obj_data_type={self.obj_data_type}, graph={self.graph}>"
        )

    def __eq__(self, other):
        return self.sub == other.sub and self.pred == other.pred and \
            self.obj == other.obj and self.graph == other.graph and \
            self.sub_type == other.sub_type and self.obj_type == other.obj_type \
            and self.obj_data_type == other.obj_data_type

    def __str__(self) -> str:
        return (
            f"<Quad sub={self.sub}, pred={self.pred}, "
            f"obj={self.obj}, graph={self.graph} "
            f"sub_type={self.sub_type}, obj_type={self.obj_type}, "
            f"obj_data_type={self.obj_data_type}>"
        )

    def __hash__(self):
        return hash((
            self.sub, self.sub_type, self.pred,
            self.obj, self.obj_type, self.obj_data_type,
            self.graph
        ))

    def __repr__(self):
        return str(self)

    def __lt__(self, other):
        return str(self.verbosify()) < str(other.verbosify())


class QuadStore:
    """Holds a set of quads

    This class is for interactions with Anzo that require trig.
    It holds quads which can be serialized to json and used
    in a semantic service payload, or it can be used to capture
    the results of a construct query, or get, for example.

    Attributes:
        quads: A set of Quads

    Raises:
        RuntimeError: Calling the method from_json errors
            when the json that it is passed is not valid.
    """

    def __init__(self, quads=None) -> None:
        """Stores quads.

        This initialization sets up an empty set of
        of Quads that can be added to by other methods.

        """

        self.quads: Set[Quad] = quads if quads else set()
        self.namespace: Dict[str, str] = NAMESPACE

    @staticmethod
    def from_anzo_json_list(json_list: List[Dict[str, Any]]): # -> QuadStore:
        """Builds a QuadStore from a list of JSON objects.

        Iterates through the list of JSON objects, and for each
        one adds a coresponding Quad to the QuadStore.

        Args:
            json_list: A list of JSON objects.

        Returns: A QuadStore object

        Raises:
            RuntimeError: If the passed JSON objects do not have the following
            form: {
                   "namedGraphUri": <URI_1> ,
                   "subject":{
                       "objectType":<TYPE>,
                       "value":<URI_2>
                       },
                  "predicate":<URI_3>,
                  "object":{
                      "objectType":<TYPE>,
                      "value":<URI_4>
                      }
                  }
        """
        quads = set()
        for item in json_list:
            quads.add(Quad.from_anzo_json_dict(item))
        return QuadStore(quads)

    @staticmethod
    def from_rdflib_graph(rdflib_graph, include_graphs: bool = True):
        """Builds a QuadStore from an RDFLib Graph

        This method requires RDFLib to be installed. If RDFLib is not
        installed, a NotImplementedError is raised.

        Args:
            rdflib_graph: RdfLib Graph to use for building the QuadStore
            include_graphs (optional): If False, then the Graph of all Quads
                is set to None, otherwise, the value in the rdflib_graph
                object is used. Defaults to True.

        Returns: A QuadStore object
        """

        quads = set()
        for s in rdflib_graph.quads():

            if isinstance(s[0], rdflib.BNode):
                sub_type = "bnode"
            else:
                sub_type = "uri"

            obj_type = ""
            obj_data_type = ""

            if isinstance(s[2], rdflib.Literal) and s[2].datatype is not None:
                obj_type = "literal"
                obj_data_type = str(s[2].datatype)
            elif isinstance(s[2], rdflib.Literal) and s[2].datatype is None:
                obj_type = "literal"
            elif isinstance(s[2], rdflib.BNode):
                obj_type = "bnode"
            else:
                obj_type = "uri"

            if include_graphs:
                graph = str(s[3].identifier)
            else:
                graph = None

            quads.add(Quad(
                sub=str(s[0]),
                pred=str(s[1]),
                obj=str(s[2]),
                graph=graph,
                sub_type=sub_type,
                obj_type=obj_type,
                obj_data_type=obj_data_type
            ))

        return QuadStore(quads)

    @staticmethod
    def from_trig_file(filename: str) -> None:
        """Builds a QuadStore from a trig file

        This method requires RDFLib to be installed. If RDFLib is not
        installed, a NotImplementedError is raised.

        Args:
            filename: Name of the file to parse
        """

        return QuadStore.from_file(filename, "trig")

    @staticmethod
    def from_file(filename: str, file_format: str):
        """Builds a QuadStore from a file. If the file format must be
        specified. The file formats that are supported are those
        that are supported by rdflib, and additional file formats
        can be supported by installing additional rdflib serializers.

        This method requires RDFLib to be installed. If RDFLib is not
        installed, a NotImplementedError is raised.

        If RDFLib is installed, then supported file_format values are
        "trig", "trix", "ttl", and "n3". If RDFLib-JSONLD is
        installed, then "json-ld" is also valid.

        Args:
            filename: Name of the file to parse
            file_format: Format of the file to parse (e.g., trig)
        """

        g = rdflib.ConjunctiveGraph()
        g.parse(filename, format=file_format)

        without_graph_file_formats = ["ttl", "n3", "trix"]
        include_graphs = file_format not in without_graph_file_formats

        return QuadStore.from_rdflib_graph(g, include_graphs)

    def as_anzo_json_list(self) -> List:
        """Returns the quad store encoded as a list of statements, adhering
        to the Anzo statement json spec.
        """
        return [quad.as_anzo_json_dict() for quad in self.quads]

    def as_list(self) -> List:
        return list(self.quads)

    def as_rdflib_graph(self):
        """
        Returns the QuadStore as an RDFLIB Conjuctive Graph

        This method requires RDFLib to be installed. If RDFLib is not
        installed, a NotImplementedError is raised.
        """

        g = rdflib.ConjunctiveGraph()
        for prefix in self.namespace:
            g.namespace_manager.bind(prefix, rdflib.URIRef(self.namespace[prefix]))

        for stmt in self:

            # Create the rdflib subject. It depends on the type of the
            # type of the object as specified in the statement
            if stmt.sub_type == "uri":
                rdflib_sub = rdflib.URIRef(stmt.sub)
            elif stmt.sub_type == "bnode":
                rdflib_sub = rdflib.BNode(stmt.sub)
            else:
                raise ValueError(f"Invalid subject type: {stmt.sub_type}")

            # Create the rdflib object. It depends on the type of the
            # type of the object as specified in the statement
            if stmt.obj_type == "uri":
                rdflib_obj = rdflib.URIRef(stmt.obj)
            elif stmt.obj_type == "bnode":
                rdflib_obj = rdflib.BNode(stmt.obj)
            elif stmt.obj_type == "":
                rdflib_obj = rdflib.Literal(stmt.obj)
            elif stmt.obj_type == "literal" and stmt.obj_data_type:
                rdflib_obj = rdflib.Literal(stmt.obj,
                                            datatype=stmt.obj_data_type)
            else:
                rdflib_obj = rdflib.Literal(stmt.obj)

            g.add((
                rdflib_sub,
                rdflib.URIRef(stmt.pred),
                rdflib_obj,
                rdflib.URIRef(stmt.graph)
            ))
        return g

    def as_record_dictionaries(self) -> List[dict]:
        """
        Returns the quad store as a list of dictionaries.
        The return format is well suited to interface with
        pandas through the DataFrame from_dict or from_records methods.

        Returns:
            A list of dictionaries, each representing one quad.
        """

        return [quad.as_record_dict() for quad in self.quads]

    def write_to_file(self, filename: str, file_format: str) -> None:
        """
        Writes the QuadStore to a file in the specified format.

        This method requires RDFLib to be installed. If RDFLib is not
        installed, a NotImplementedError is raised.

        If RDFLib is installed, then supported file_format values are
        "trig", "trix", "ttl", and "n3". If RDFLib-JSONLD is
        installed, then "json-ld" is also valid.

        Args:
            filename: Name of the file to write to
            file_format: Format of the file (e.g., "trig")
        """

        self.as_rdflib_graph().serialize(
            destination=filename, format=file_format
        )

    def write_to_trig_file(self, filename: str) -> None:
        """
        Writes the QuadStore to a trig file.

        This method requires RDFLib to be installed. If RDFLib is not
        installed, a NotImplementedError is raised.

        Args:
            filename: Name of the file to write to
        """

        self.write_to_file(filename, "trig")

    @staticmethod
    def from_string(data: str, string_format: str) -> str:
        g = rdflib.ConjunctiveGraph()
        g.parse(data=data, format=string_format)
        without_graph_formats = ["ttl", "n3", "trix"]
        include_graphs = string_format not in without_graph_formats
        return QuadStore.from_rdflib_graph(g, include_graphs)

    @staticmethod
    def from_trig_string(data: str) -> str:
        return QuadStore.from_string(data, "trig")

    def add_prefixes(self, data: dict) -> dict:
        """
        Adds prefixes to the dict of prefixes in a QuadStore

        Args:
            data: A dictionary of prefixes and uris to add
        """
        for key, val in data.items():
            self.namespace[key] = val
        return self.namespace

    def to_string(self, string_format: str) -> str:
        return self.as_rdflib_graph().serialize(
            format=string_format
        ).decode()

    def to_trig_string(self) -> str:
        return self.to_string("trig")

    def filter(self,
               subject="",
               predicate="",
               statement_object="",
               graph=""):  # -> QuadStore:
        """Filters QuadStore on Search Terms.

        This generates a new QuadStore where the
        quads in that new store match the search
        terms provided conjunctively. Failing to
        provide a search term, or setting it to
        the empty string treats that as a wildcard
        search term. The filtering works as a
        series of gates on each quad in the store.
        Any quad that does not pass one of the
        gates will not be included in the result.

        Args:
            subject: A string. The subject that all quads in the result have.
            predicate: A string. The predicate that
                all quads in the result have.
            statement_object: A string. The object
                all quads in the result have.
            graph: A string. The graph all quads in the result have.

        Returns: A QuadStore that is the result of
            filtering conjunctively by all the args
            that were passed to it.
        """
        result = QuadStore()

        def check_match(filter_arg, quad_item):
            match = True
            if filter_arg:
                if not filter_arg == quad_item:
                    match = False
            return match

        for quad in self.quads:
            subject_matches = check_match(subject, quad.sub)
            if not subject_matches:
                continue
            predicate_matches = check_match(predicate, quad.pred)
            if not predicate_matches:
                continue
            object_matches = check_match(statement_object, quad.obj)
            if not object_matches:
                continue
            graph_matches = check_match(graph, quad.graph)
            if not graph_matches:
                continue
            result.quads.add(quad)
        return result

    def add(self, quad: Quad) -> bool:
        """ Add the given quad to a quad store.

        Args:
            quad: the quad to add

        Returns: True if the quad store did not already contain the given quad
        """
        if quad not in self.quads:
            self.quads.add(quad)
            return True
        return False

    def remove(self, quad: Quad) -> bool:
        """ Remove the given quad from a quad store.

        Args:
            quad: the quad to remove

        Returns: A boolean indicating whether the quad was found and removed
        """
        if quad in self.quads:
            self.quads.remove(quad)
            return True
        return False

    def set(self, quad: Quad) -> bool:
        """ Convenience method to "upsert" the given quad to the quad store.

        First removes any existing quads from the quad store that have the same
        graph/subject/predicate as the given quad.
        Then takes the given quad and add it to the quad store.

        Based on:
        https://rdflib.readthedocs.io/en/stable/apidocs/rdflib.html#rdflib.graph.Graph.set

        Args:
            quad: the quad to upsert
        Return:
            A boolean indicating whether any quads were removed
        """
        to_remove = self.filter(subject=quad.sub,
                                predicate=quad.pred, graph=quad.graph)
        for remove in to_remove:
            self.remove(remove)
        self.add(quad)
        return len(to_remove) > 0

    def prettify(self) -> str:
        str_list = []
        for quad in self.quads:
            str_list.append(str(quad))
        return "\n".join(str_list)

    def verbosify(self) -> str:
        str_list = []
        for quad in self.quads:
            str_list.append(quad.verbosify())
        return "\n".join(str_list)

    def __eq__(self, other_store) -> bool:
        return self.quads == other_store.quads

    def __iter__(self):
        return iter(self.quads)

    def __len__(self):
        return len(self.quads)



# /*******************************************************************************
#  * Copyright (c) 2019 - 2022 Cambridge Semantics Incorporated.
#  * All rights reserved.
#  * 
#  * Contributors:
#  *     Cambridge Semantics Incorporated
#  *******************************************************************************/

import parameterized
import unittest

from pyanzo import (
    AnzoClient,
    QueryResult,
    TableResult,
)

from .test_common import (
    GRAPHMART,
    SERVER,
    PORT,
    USERNAME,
    PASSWORD
)


class TestIntegratedQueryFuzz(unittest.TestCase):
    graphmart = GRAPHMART

    # Templated verions of the queries to allow for different SELECT vars and LIMITs  # noqa
    simple_select_query_template = """
        PREFIX data: <http://cambridgesemantics.com/ont/autogen/a69a/PyAnzo_Dictionary/PyAnzo_Datasource#>

        SELECT {} WHERE {{
            ?s a data:Person ;
                data:Person_City ?city .
        }} ORDER BY ?city LIMIT {}
    """  # noqa

    query_with_optional_template = """
        PREFIX data: <http://cambridgesemantics.com/ont/autogen/a69a/PyAnzo_Dictionary/PyAnzo_Datasource#>

        SELECT {} WHERE {{
          ?s a data:Person ;
             data:Person_Name ?name ;
             data:Person_City ?city ;
             data:Person_Age ?age .

          # person height is not a property and
          # will be unbound by design (for testing)
          OPTIONAL {{ ?s  data:Person_Height  ?height }}

            }} ORDER BY ?city LIMIT {}
    """  # noqa

    QueryResult_select_star = QueryResult("""{
        "head": {"vars": ["s", "name", "city", "age", "height"]},
        "results":
            {"bindings": [
                {"age": {"datatype": "http://www.w3.org/2001/XMLSchema#long",
                         "type": "literal",
                         "value": "60"},
                 "city": {"type": "literal", "value": "Boston"},
                 "name": {"type": "literal", "value": "Diego"},
                 "s": {"type": "uri",
                       "value": "http://csi.com/Person/879d629b-1015-45b9-8457-7271c0f2a9f0"}
                },
                {"age": {"datatype": "http://www.w3.org/2001/XMLSchema#long",
                         "type": "literal",
                         "value": "33"},
                 "city": {"type": "literal", "value": "Cambridge"},
                 "name": {"type": "literal", "value": "Bob"},
                 "s": {"type": "uri",
                       "value": "http://csi.com/Person/c2ee95a3-d5b1-40d8-8243-9c7d541853f7"}
                },
                {"age": {"datatype": "http://www.w3.org/2001/XMLSchema#long",
                         "type": "literal",
                         "value": "70"},
                 "city": {"type": "literal", "value": "Medford"},
                 "name": {"type": "literal", "value": "Eve"},
                 "s": {"type": "uri",
                       "value": "http://csi.com/Person/ffbe9982-7491-42da-9398-e4c22fab5711"}
                },
                {"age": {"datatype": "http://www.w3.org/2001/XMLSchema#int",
                         "type": "literal",
                         "value": "30"},
                 "city": {"type": "literal", "value": "Somerville"},
                 "name": {"type": "literal", "value": "Alice"},
                 "s": {"type": "uri",
                       "value": "http://csi.com/Person/a153b0aa-5583-40ff-8710-f94c8a25df3d"}
                },
                {"age": {"datatype": "http://www.w3.org/2001/XMLSchema#long",
                         "type": "literal",
                         "value": "50"},
                 "city": {"type": "literal", "value": "Watertown"},
                 "name": {"type": "literal", "value": "Carlos"},
                 "s": {"type": "uri",
                       "value": "http://csi.com/Person/0c391c0f-340d-4f08-9fa6-13a1793c35cb"}
                }
            ]
           }
        }""") # noqa

    # The test docstrings are convenient when looking at the code, but
    # they make the verbose testing view look cluttered. This function
    # removes comments/descriptions from the verbose test view.
    def shortDescription(self):
        return None

    def setUp(self) -> None:
        self.anzo_client = AnzoClient(
            server=SERVER, username=USERNAME,
            password=PASSWORD, port=PORT
        )

    @parameterized.parameterized.expand([
        ("SELECT_CITY_HEIGHT", "?city ?height", ["city", "height"]),
        ("SELECT_*", "*", ["*"]),
        ("SELECT_CITY", "?city", ["city"]),
        ("SELECT_CITY_AGE_NAME", "?city ?age ?name", ["city", "name", "age"])
    ])
    def test_graphmart_query_various_selects(self,
                                             test_name,
                                             select_vars,
                                             select_vars_list):
        """Test querying a graphmart with different combinations of
           select variables"""

        # attempts a query based on the specified select vars
        query_string = self.query_with_optional_template.format(select_vars, 5)

        result = self.anzo_client.query_graphmart(
            self.graphmart, query_string=query_string
        ).as_table_results()

        self.assertEqual(
            result.as_record_dictionaries(),

            # QueryResult_select_star is the result of selecting all variables
            # Here we are filtering out the vars not in our select in order to
            #   check for equality
            [
                {key: val for key, val in solution_dict.items()
                    if key in select_vars_list or select_vars == "*"}
                for solution_dict in self.QueryResult_select_star.as_table_results().as_record_dictionaries()
            ]
        )

    @parameterized.parameterized.expand([
        ("LIST", TableResult.as_list),
        ("DICT", TableResult.as_record_dictionaries),
        # Need to convert iter to list in order to compare results
        ("LIST_ITER", lambda x: list(TableResult.as_list_iter(x)))
    ])
    def test_graphmart_query_result_length(self,
                                           test_name,
                                           query_result_method) -> None:
        """Test that the length of the query result is current (<= LIMIT)"""

        for limit in range(5):
            query_string = self.simple_select_query_template.format("*", limit)
            result = self.anzo_client.query_graphmart(
                self.graphmart, query_string=query_string
            ).as_table_results()

            self.assertLessEqual(len(query_result_method(result)), limit)



# /*******************************************************************************
#  * Copyright (c) 2019 - 2022 Cambridge Semantics Incorporated.
#  * All rights reserved.
#  * 
#  * Contributors:
#  *     Cambridge Semantics Incorporated
#  *******************************************************************************/

import unittest

from pyanzo import AnzoClient
from pyanzo.graphmart_manager import GraphmartManager

from .test_common import (
    GRAPHMART,
    SERVER,
    PORT,
    USERNAME,
    PASSWORD,
    ANZOGRAPH_DS
)


class TestGraphmartManager(unittest.TestCase):
    # Paramterized values for the tests. This value is overwritten by the
    # values in the decorator

    graphmart = GRAPHMART
    anzograph_ds = ANZOGRAPH_DS

    # A simple select query used for testing core functionality in the tests
    select_query = """
        PREFIX data: <http://cambridgesemantics.com/ont/autogen/a69a/PyAnzo_Dictionary/PyAnzo_Datasource#>

        SELECT ?city WHERE {
            ?s a data:Person ;
                data:Person_City ?city .
        } ORDER BY ?city LIMIT 5
    """  # noqa

    select_query_results_first_layer = [
        ["Cambridge"],
        ["Somerville"],
        ["Watertown"],
    ]

    select_query_results_second_layer = [
        ["Boston"],
        ["Medford"],
    ]

    select_query_results = [
        ["Boston"],
        ["Cambridge"],
        ["Medford"],
        ["Somerville"],
        ["Watertown"],
    ]

    layer_one_uri = "http://cambridgesemantics.com/Layer/dd29f470d8d44dad8e6cf9ce3b6322fd"  # noqa
    layer_two_uri = "http://cambridgesemantics.com/Layer/d2bb3f418d7942d8a85fdedad608ccb3"  # noqa

    load_step = "http://cambridgesemantics.com/LoadDataStep/1347736407234bdc84f7b04991c70541"  # noqa
    query_step = "http://cambridgesemantics.com/QueryStep/390e0618d5d04dd5ac57cbf26947f397"  # noqa

    @classmethod
    def setUpClass(cls) -> None:
        ac = AnzoClient(
            server=SERVER, username=USERNAME, password=PASSWORD, port=PORT
        )
        gm = GraphmartManager(ac, cls.graphmart)

        # Make sure the graphmart is started in the right state
        gm.enable_layers()
        gm.activate()
        gm.reload()

    @classmethod
    def tearDownClass(cls) -> None:
        ac = AnzoClient(
            server=SERVER, username=USERNAME, password=PASSWORD, port=PORT
        )
        gm = GraphmartManager(ac, cls.graphmart)

        # Make sure the graphmart is left in the right state
        gm.enable_layers()
        gm.reload()
        gm.activate()

    def setUp(self) -> None:
        self.anzo_client = AnzoClient(
            server=SERVER, username=USERNAME, password=PASSWORD, port=PORT
        )

        self.graphmart_manager = GraphmartManager(
            self.anzo_client, self.graphmart
        )

    def test_query(self) -> None:
        result = self.graphmart_manager.query(
            query_string=self.select_query
        )

        self.assertEqual(
            result.as_table_results().as_list(), self.select_query_results
        )

    def test_enable_disable_layers(self) -> None:
        # First, disenable the first layer, and confirm things look good
        self.graphmart_manager.disable_layers([self.layer_one_uri])
        self.graphmart_manager.reload()
        self.assertTrue(self.graphmart_manager.is_graphmart_online())

        self.assertFalse(
            self.graphmart_manager.are_layers_online([self.layer_one_uri])
        )

        self.assertTrue(
            self.graphmart_manager.are_layers_online([self.layer_two_uri])
        )

        result = self.graphmart_manager.query(self.select_query)
        self.assertEqual(
            result.as_table_results().as_list(),
            self.select_query_results_second_layer
        )

        # Second, re-enable, and confirm things look good
        self.graphmart_manager.enable_layers([self.layer_one_uri])
        self.graphmart_manager.reload()
        self.assertTrue(self.graphmart_manager.is_graphmart_online())

        self.assertTrue(
            self.graphmart_manager.are_layers_online([self.layer_one_uri])
        )

        self.assertTrue(
            self.graphmart_manager.are_layers_online([self.layer_two_uri])
        )
        self.assertTrue(self.graphmart_manager.is_graphmart_online())

        result = self.graphmart_manager.query(self.select_query)
        self.assertEqual(
            result.as_table_results().as_list(), self.select_query_results
        )

    def test_enable_disable_steps(self) -> None:
        load_steps = [self.load_step]
        query_steps = [self.query_step]

        self.graphmart_manager.enable_steps(query_steps)
        self.graphmart_manager.refresh()
        self.assertTrue(self.graphmart_manager.are_steps_online(query_steps))
        self.assertTrue(self.graphmart_manager.are_steps_online(load_steps))

        self.graphmart_manager.disable_steps(query_steps)
        self.graphmart_manager.refresh()
        self.assertFalse(self.graphmart_manager.are_steps_online(query_steps))
        self.assertTrue(self.graphmart_manager.are_steps_online(load_steps))

        self.graphmart_manager.enable_steps(query_steps)
        self.graphmart_manager.refresh()
        self.assertTrue(self.graphmart_manager.are_steps_online(query_steps))
        self.assertTrue(self.graphmart_manager.are_steps_online(load_steps))

    def test_reload(self) -> None:
        self.graphmart_manager.reload()

    def test_refresh(self) -> None:
        self.graphmart_manager.reload()

    def test_are_layers_online_with_args(self) -> None:
        self.assertTrue(self.graphmart_manager.are_layers_online(
            [self.layer_one_uri, self.layer_two_uri]
        ))

    def test_are_layers_online_without_args(self) -> None:
        self.assertTrue(self.graphmart_manager.are_layers_online())

    def test_get_layers(self) -> None:
        self.assertEqual(
            sorted(self.graphmart_manager.get_layers()),
            sorted([self.layer_one_uri, self.layer_two_uri])
        )

    def test_are_steps_online_with_args(self) -> None:
        self.assertTrue(self.graphmart_manager.are_steps_online(
            [self.load_step, self.query_step]
        ))

    def test_are_steps_online_without_args(self) -> None:
        self.assertTrue(self.graphmart_manager.are_steps_online())

    def test_get_steps(self) -> None:
        self.assertEqual(
            sorted(self.graphmart_manager.get_steps()),
            sorted([self.load_step, self.query_step])
        )

    def test_is_online(self) -> None:
        self.assertTrue(self.graphmart_manager.is_graphmart_online())

    def test_deactivate_and_activate_without_azg_uri(self) -> None:
        self.assertTrue(self.graphmart_manager.are_layers_online())
        self.assertTrue(self.graphmart_manager.is_graphmart_online())

        self.graphmart_manager.deactivate()

        self.assertFalse(self.graphmart_manager.is_graphmart_online())
        self.assertFalse(self.graphmart_manager.are_layers_online())

        self.graphmart_manager.activate()

        self.assertTrue(self.graphmart_manager.are_layers_online())
        self.assertTrue(self.graphmart_manager.is_graphmart_online())

    def test_deactivate_and_activate_with_azg_uri(self) -> None:
        self.assertTrue(self.graphmart_manager.are_layers_online())
        self.assertTrue(self.graphmart_manager.is_graphmart_online())

        self.graphmart_manager.deactivate()

        self.assertFalse(self.graphmart_manager.is_graphmart_online())
        self.assertFalse(self.graphmart_manager.are_layers_online())

        self.graphmart_manager.activate(self.anzograph_ds)

        self.assertTrue(self.graphmart_manager.are_layers_online())
        self.assertTrue(self.graphmart_manager.is_graphmart_online())

    def test_activating_twice(self) -> None:
        self.graphmart_manager.activate()
        self.graphmart_manager.activate()
        self.assertTrue(self.graphmart_manager.is_graphmart_online())

    def test_deactivating_twice(self) -> None:
        self.graphmart_manager.deactivate()
        self.graphmart_manager.deactivate()
        self.assertFalse(self.graphmart_manager.is_graphmart_online())

        # put graphmart back in its normal state
        self.graphmart_manager.activate()
        self.assertTrue(self.graphmart_manager.is_graphmart_online())


    def test_dynamic_activate(self) -> None:
        #dynamic_activate(self, azg_launch_config: str = "", azg_version:str = "", azg_size: str = "") 
        
        self.graphmart_manager.dynamic_activate(self.anzograph_ds)
        self.assertTrue(self.graphmart_manager.are_layers_online())
        self.assertTrue(self.graphmart_manager.is_graphmart_online())




# /*******************************************************************************
#  * Copyright (c) 2019 - 2022 Cambridge Semantics Incorporated.
#  * All rights reserved.
#  * 
#  * Contributors:
#  *     Cambridge Semantics Incorporated
#  *******************************************************************************/

import unittest

from pyanzo import (
    AnzoClient, ALL_NAMED_GRAPHS_URI
)

from .test_common import (
    SERVER,
    PORT,
    USERNAME,
    PASSWORD
)


class TestAnzoClientQueryLds(unittest.TestCase):
    ontology_lds_uri = "http://openanzo.org/catEntry(%5Bhttp%3A%2F%2Fcambridgesemantics.com%2Fregistries%2FOntologies%5D%40%5Bhttp%3A%2F%2Fopenanzo.org%2Fdatasource%2FsystemDatasource%5D)"  # noqa

    pyanzo_ontology_uri = "http://cambridgesemantics.com/ont/autogen/a69a/PyAnzo_Dictionary/PyAnzo_Datasource"  # noqa
    resources_in_pyanzo_ontology = [
        ["http://cambridgesemantics.com/ont/autogen/a69a/PyAnzo_Dictionary/PyAnzo_Datasource#Person"],  # noqa
        ["http://cambridgesemantics.com/ont/autogen/a69a/PyAnzo_Dictionary/PyAnzo_Datasource#Person_Age"],  # noqa
        ["http://cambridgesemantics.com/ont/autogen/a69a/PyAnzo_Dictionary/PyAnzo_Datasource#Person_City"],  # noqa
        ["http://cambridgesemantics.com/ont/autogen/a69a/PyAnzo_Dictionary/PyAnzo_Datasource#Person_Name"]  # noqa
     ]

    pyanzo_ont_query_string = f"""
        PREFIX owl: <http://www.w3.org/2002/07/owl#>

        SELECT distinct ?resource WHERE {{
            GRAPH <{pyanzo_ontology_uri}> {{
                ?resource a ?type .
                VALUES (?type) {{
                    (owl:Class)
                    (owl:DatatypeProperty)
                    (owl:FunctionalProperty)
                }}
            }}
        }}
        ORDER BY ?resource
        LIMIT 20
    """  # noqa

    # This query is designed to make sure that when (a) no named graphs are
    # specifed and (b) there's no GRAPH ?g part of the where then the query
    # query will still reach the ontology graphs.
    pyanzo_ont_query_string_without_graph2 = f"""
        PREFIX owl: <http://www.w3.org/2002/07/owl#>
        PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>

        SELECT distinct ?resource WHERE {{
            ?resource a owl:Class .
            ?resource rdfs:label "Person" .
        }}
        ORDER BY ?resource
        LIMIT 20
    """  # noqa

    person_resources = [
        ['http://cambridgesemantics.com/ont/autogen/a69a/PyAnzo_Dictionary/PyAnzo_Datasource#Person'],
        ['http://s.opencalais.com/1/type/em/e/Person'],
        ['http://xmlns.com/foaf/0.1/Person']
    ]


    system_table_lds = "http://openanzo.org/catEntry(%5Bhttp%3A%2F%2Fcambridgesemantics.com%2Fontologies%2F2009%2F05%2FLinkedData%23AnzoXray%5D%40%5Bhttp%3A%2F%2Fcambridgesemantics.com%2Fdatasource%2FSystemTables%5D)"  # noqa

    gmart_query = """
        PREFIX gms: <http://cambridgesemantics.com/ontologies/GraphmartStatus#>
        SELECT ?status WHERE {
            ?obj gms:status ?status .
            FILTER(?obj=<http://cambridgesemantics.com/Graphmart/9da211618a15476daa10cead2292d8e7>)
        }
        LIMIT 100
    """  # noqa

    gmart_query_res = [
        ['http://openanzo.org/ontologies/2008/07/System#Online']
    ]

    def setUp(self) -> None:
        self.anzo_client = AnzoClient(
            server=SERVER, username=USERNAME,
            password=PASSWORD, port=PORT
        )

    def test_bad_query(self) -> None:
        query_string = "SELECT * WHER { ?s ?p ?o . } LIMIT 10"

        self.assertRaises(
            RuntimeError, self.anzo_client.query_lds,
            self.ontology_lds_uri, query_string=query_string
        )

    def test_empty_query(self) -> None:
        query_string = ""
        self.assertRaises(
            ValueError, self.anzo_client.query_lds,
            self.ontology_lds_uri, query_string=query_string
        )

    def test_bad_lds_uri(self) -> None:
        # make bad by dropping the last two characters of the
        # ontology lds uri
        bad_lds_uri = self.ontology_lds_uri[:-2]

        self.assertRaises(
            RuntimeError, self.anzo_client.query_lds,
            bad_lds_uri, query_string=self.pyanzo_ont_query_string
        )

    def test_empty_lds_uri(self) -> None:
        empty_lds_uri = ""

        self.assertRaises(
            RuntimeError, self.anzo_client.query_lds,
            empty_lds_uri, query_string=self.pyanzo_ont_query_string
        )

    def test_simple_lds_query(self) -> None:
        result = self.anzo_client.query_lds(
            self.ontology_lds_uri, query_string=self.pyanzo_ont_query_string
        )

        self.assertEqual(result.as_table_results().as_list(),
                         self.resources_in_pyanzo_ontology)

    def test_query_without_named_graph_specified(self) -> None:
        # test query is designed to make sure that when (a) no named graphs are
        # specifed in the query and (b) there's no GRAPH ?g part of the where
        # then the query query will still reach the ontology graphs.

        result = self.anzo_client.query_lds(
            self.ontology_lds_uri,
            query_string=self.pyanzo_ont_query_string_without_graph2
        )

        self.assertEqual(
            result.as_table_results().as_list(), self.person_resources
        )

    def test_query_system_tables_with_all_graphs(self) -> None:
        # When querying the system tables, all named graphs should be included
        res = self.anzo_client.query_lds(
            self.system_table_lds, self.gmart_query,
            named_graphs=[ALL_NAMED_GRAPHS_URI]
        )
        res_list = res.as_table_results().as_list()
        self.assertEqual(res_list, self.gmart_query_res)



# /*******************************************************************************
#  * Copyright (c) 2019 - 2022 Cambridge Semantics Incorporated.
#  * All rights reserved.
#  * 
#  * Contributors:
#  *     Cambridge Semantics Incorporated
#  *******************************************************************************/

TYPE_PRED = "http://www.w3.org/1999/02/22-rdf-syntax-ns#type"
ALL_NAMED_GRAPHS_URI = "http://openanzo.org/namedGraphs/reserved/graphs/ALL"
SYSTEM_DATASOURCE_URI = "http://openanzo.org/datasource/systemDatasource"

NAMESPACE = {
    "anzo":             "http://openanzo.org/ontologies/2008/07/Anzo#",
    "cs-ds":            "http://cambridgesemantics.com/datasets/",
    "cs-ld":            "http://cambridgesemantics.com/ontologies/2009/05/LinkedData#",
    "dc":               "http://purl.org/dc/elements/1.1/",
    "du-ls":            "http://openanzo.org/ontologies/2008/07/ASDL/IngestManager/DistributedUnstructuredLoadService#",
    "etl-ls":           "http://openanzo.org/ontologies/2008/07/ASDL/IngestManager/EtlLoadService#",
    "gm-ls":            "http://openanzo.org/ontologies/2008/07/ASDL/IngestManager/GraphmartLoadService#",
    "im":               "http://openanzo.org/ontologies/2008/07/ASDL/IngestManager#",
    "lds":              "http://cambridgesemantics.com/LinkedDataSet/",
    "orch":             "http://openanzo.org/ontologies/2008/07/ASDL/Orchestration#",
    "reg":              "http://cambridgesemantics.com/registries/",
    "registries":       "http://openanzo.org/registries/",
    "remoteconnection": "http://openanzo.org/ontologies/2008/07/RemoteConnection#",
    "role":             "http://openanzo.org/Role/",
    "ss":               "http://openanzo.org/ontologies/2008/07/SemanticService#",
    "svc":              "http://cambridgesemantics.com/service/",
    "system":           "http://openanzo.org/ontologies/2008/07/System#"
}

DNG_PRED = NAMESPACE["anzo"] + "defaultNamedGraph"
NG_TYPE = NAMESPACE["anzo"] + "NamedGraph"



# /*******************************************************************************
#  * Copyright (c) 2019 - 2022 Cambridge Semantics Incorporated.
#  * All rights reserved.
#  * 
#  * Contributors:
#  *     Cambridge Semantics Incorporated
#  *******************************************************************************/

import sys
import logging

# See readme for information on how to setup Anzo for tests
GRAPHMART = "http://cambridgesemantics.com/Graphmart/9da211618a15476daa10cead2292d8e7"  # noqa
ANZOGRAPH_DS = "http://cambridgesemantics.com/GqeDatasource/guid_3af6a337ad444400a1f01dbe2e16b82b"  # noqa

SERVER = "localhost"
PORT = "8443"
AUTH_TOKEN = "auth_token"
EMPTY_AUTH_TOKEN = ""
USERNAME = "sysadmin"
PASSWORD = "123"

def log_while_running_tests():
    logger = logging.getLogger()
    logger.level = logging.DEBUG
    stream_handler = logging.StreamHandler(sys.stdout)
    logger.addHandler(stream_handler)




# /*******************************************************************************
#  * Copyright (c) 2019 - 2022 Cambridge Semantics Incorporated.
#  * All rights reserved.
#  * 
#  * Contributors:
#  *     Cambridge Semantics Incorporated
#  *******************************************************************************/

import unittest
from pyanzo import AnzoClient, Quad, QuadStore

from .test_common import (
    SERVER,
    PORT,
    USERNAME,
    PASSWORD,
)


class TestAnzoClientAddAndRemove(unittest.TestCase):
    named_graph_uri = "http://cambridgesemantics.com/namedgraph/addremove"
    non_existent_named_graph_uri = "http://cambridgesemantics.com/namedgraph/doesntexist"  # noqa

    quad1 = Quad(
        sub="http://cambridgesemantics.com/namedgraph/addremove/sub1",
        pred="http://cambridgesemantics.com/namedgraph/addremove/pred2",
        obj="http://cambridgesemantics.com/namedgraph/addremove/obj1",
        graph=named_graph_uri
    )

    quad2 = Quad(
        sub="http://cambridgesemantics.com/namedgraph/addremove/sub1",
        pred="http://cambridgesemantics.com/namedgraph/addremove/pred2",
        obj="http://cambridgesemantics.com/namedgraph/addremove/obj2",
        graph=named_graph_uri
    )

    quad3 = Quad(
        sub="http://cambridgesemantics.com/namedgraph/addremove/sub1",
        pred="http://cambridgesemantics.com/namedgraph/addremove/pred3",
        obj="http://cambridgesemantics.com/namedgraph/addremove/obj2",
        graph=named_graph_uri
    )

    non_existent_quad = Quad(
        sub="http://cambridgesemantics.com/namedgraph/addremove/sub1",
        pred="http://cambridgesemantics.com/namedgraph/addremove/pred2",
        obj="http://cambridgesemantics.com/namedgraph/addremove/obj1",
        graph=non_existent_named_graph_uri
    )

    sample_prefix = "http://csi.test.com/"

    sample_quad1 = Quad(
        sub=f"{sample_prefix}subject1",
        pred=f"{sample_prefix}predicate1",
        obj=f"{sample_prefix}object1",
        graph=sample_prefix)

    sample_quad2 = Quad(
        sub=f"{sample_prefix}subject2",
        pred=f"{sample_prefix}predicate2",
        obj=f"{sample_prefix}object2",
        graph=sample_prefix)

    sample_graph1 = QuadStore(set([sample_quad1, sample_quad2]))

    sample_graph2 = QuadStore(set([Quad(
        sub=f"{sample_prefix}subject2",
        pred=f"{sample_prefix}predicate3",
        obj="7",
        graph=f"{sample_prefix}graph2",
        obj_type="literal",
        obj_data_type="http://www.w3.org/2001/XMLSchema#integer")]))

    @classmethod
    def wipe_graph(cls) -> None:
        ac = AnzoClient(SERVER, PORT, username=USERNAME, password=PASSWORD)
        query = f"""
            DELETE {{ GRAPH ?g {{ ?s ?p ?o }} }}
            WHERE {{
                GRAPH ?g {{ ?s ?p ?o }}
                FILTER(?g=?graph)
                VALUES(?graph)
                {{(<{cls.named_graph_uri}>)
                  (<{cls.sample_prefix}>)
                  (<{cls.sample_prefix}graph2>)
                  }}
            }}
        """
        ac.update_journal(query)

    @classmethod
    def setUpClass(cls) -> None:
        cls.wipe_graph()

    @classmethod
    def tearDownClass(cls) -> None:
        cls.wipe_graph()

    def setUp(self) -> None:
        self.anzo_client = AnzoClient(
            SERVER, PORT, username=USERNAME, password=PASSWORD
        )

    def test_add_and_remove(self) -> None:
        self.anzo_client.add(quad_store=QuadStore({self.quad1}))
        self.assertEqual(
            QuadStore({self.quad1}),
            self.anzo_client.get(self.named_graph_uri)
        )

        self.anzo_client.add(quad_store=QuadStore({self.quad2, self.quad3}))

        self.assertEqual(
            QuadStore({self.quad1, self.quad2, self.quad3}),
            self.anzo_client.get(self.named_graph_uri)
        )

        self.anzo_client.remove(QuadStore({self.quad1}))
        self.assertEqual(
            QuadStore({self.quad2, self.quad3}),
            self.anzo_client.get(self.named_graph_uri)
        )

        self.anzo_client.remove(QuadStore({self.quad2, self.quad3}))
        self.assertEqual(
            QuadStore({}),
            self.anzo_client.get(self.named_graph_uri)
        )

    def test_add_from_file(self) -> None:

        self.anzo_client.add(trig_filename="tests/test_assets/sample.trig")
        self.assertEqual(self.sample_graph1,
                         self.anzo_client.get(self.sample_prefix))

        self.assertEqual(self.sample_graph2,
                         self.anzo_client.get(f"{self.sample_prefix}graph2"))

    def test_remove_from_file(self) -> None:

        self.anzo_client.add(quad_store=self.sample_graph1)
        self.anzo_client.add(quad_store=self.sample_graph2)

        self.assertNotEqual(
            QuadStore({}),
            self.anzo_client.get(self.sample_prefix))

        self.assertNotEqual(
            QuadStore({}),
            self.anzo_client.get(f"{self.sample_prefix}graph2"))

        self.anzo_client.remove(trig_filename="tests/test_assets/sample.trig")

        self.assertEqual(
            QuadStore({}),
            self.anzo_client.get(self.sample_prefix))

        self.assertEqual(
            QuadStore({}),
            self.anzo_client.get(f"{self.sample_prefix}graph2"))

    def test_add_and_remove_too_many_inputs(self) -> None:
        trig = "tests/test_assets/sample.trig"
        self.assertRaises(ValueError,
                          self.anzo_client.add,
                          quad_store=self.sample_graph1,
                          trig_filename=trig)

        self.assertRaises(ValueError,
                          self.anzo_client.remove,
                          quad_store=self.sample_graph1,
                          trig_filename=trig)

    def test_add_and_remove_too_few_inputs(self) -> None:
        self.assertRaises(ValueError,
                          self.anzo_client.add)

        self.assertRaises(ValueError,
                          self.anzo_client.remove)

    def test_removing_non_existent_quad(self) -> None:
        self.anzo_client.remove(QuadStore({self.non_existent_quad}))



# /*******************************************************************************
#  * Copyright (c) 2019 - 2022 Cambridge Semantics Incorporated.
#  * All rights reserved.
#  * 
#  * Contributors:
#  *     Cambridge Semantics Incorporated
#  *******************************************************************************/

import unittest
from pyanzo import AnzoClient, AnzoRequestBuilder
import requests

from .test_common import (
    SERVER,
    PORT,
    USERNAME,
    PASSWORD,
    GRAPHMART,
)

AUTH = "AUTH_TOKEN"


class TestAuthenticationClient(unittest.TestCase):
    # A simple select query used for testing core functionality in the tests
    select_query = """
        PREFIX data: <http://cambridgesemantics.com/ont/autogen/a69a/PyAnzo_Dictionary/PyAnzo_Datasource#>

        SELECT ?city WHERE {
            ?s a data:Person ;
                data:Person_City ?city .
        } ORDER BY ?city LIMIT 5
    """  # noqa

    select_query_results = [
        ["Boston"],
        ["Cambridge"],
        ["Medford"],
        ["Somerville"],
        ["Watertown"],
    ]

    def test_client_with_auth(self) -> None:
        url = f"https://{SERVER}:{PORT}/anzo_authenticate?client_name=AnzoFormClient"  # noqa

        session = requests.Session()
        data = {
            "anzo_username": USERNAME,
            "anzo_password": PASSWORD,
        }

        session.post(url, verify=False, data=data)
        half_of_token = session.cookies.get_dict()['BAYEUX_BROWSER']
        auth_token = f"BAYEUX_BROWSER={half_of_token}"

        anzo_client = AnzoClient(SERVER, PORT, auth_token=auth_token)

        res = anzo_client.query_graphmart(GRAPHMART, self.select_query)
        print(res.as_table_results().as_list())
        self.assertEqual(
            res.as_table_results().as_list(), self.select_query_results
        )

    def test_client_no_auth(self) -> None:
        self.assertRaises(ValueError, AnzoClient, SERVER, PORT)

    def test_client_username_no_password(self) -> None:
        self.assertRaises(ValueError,
                          AnzoClient,
                          SERVER,
                          PORT,
                          username=USERNAME)

    def test_client_password_no_username(self) -> None:
        self.assertRaises(ValueError,
                          AnzoClient,
                          SERVER,
                          PORT,
                          password=PASSWORD)

    def test_client_username_password_and_auth_token(self) -> None:
        self.assertRaises(ValueError,
                          AnzoClient,
                          SERVER,
                          PORT,
                          username=USERNAME,
                          password=PASSWORD,
                          auth_token=AUTH)


class TestAuthenticationErrorsRequestBuilder(unittest.TestCase):

    arb = AnzoRequestBuilder()

    def test_overloaded_builder(self) -> None:
        self.assertRaises(ValueError,
                          self.arb.with_auth,
                          USERNAME,
                          PASSWORD,
                          AUTH)

    def test_builder_username_no_password(self) -> None:
        self.assertRaises(ValueError, self.arb.with_auth, USERNAME, "", "")

    def test_builder_password_no_username(self) -> None:
        self.assertRaises(ValueError, self.arb.with_auth, "", PASSWORD, "")

    def test_builder_no_auth(self) -> None:
        self.assertRaises(ValueError, self.arb.with_auth, "", "", "")



# /*******************************************************************************
#  * Copyright (c) 2019 - 2022 Cambridge Semantics Incorporated.
#  * All rights reserved.
#  * 
#  * Contributors:
#  *     Cambridge Semantics Incorporated
#  *******************************************************************************/

import unittest
from pyanzo import AnzoClient, QuadStore, Quad
from .test_common import (
    GRAPHMART,
    SERVER,
    PORT,
    USERNAME,
    PASSWORD
)


class TestAnzoClientExecuteSemanticService(unittest.TestCase):
    reload_graphmart_service_uri = "http://cambridgesemantics.com/semanticServices/gqe#reloadGraphmart"  # noqa

    graphmart = GRAPHMART

    reload_graphmart_request = QuadStore({
        Quad("http://openanzo.org/semanticServices/",
             "http://www.w3.org/1999/02/22-rdf-syntax-ns#type",
             "http://cambridgesemantics.com/ontologies/Graphmarts#GraphmartRequest",  # noqa
             "http://openanzo.org/semanticServices/"),
        Quad("http://openanzo.org/semanticServices/",
             "http://cambridgesemantics.com/ontologies/Graphmarts#graphmart",
             graphmart,
             "http://openanzo.org/semanticServices/")
    })

    binary_service_uri = "http://openanzo.org/semanticServices/binaryStoreService#getServerConfig"  # noqa
    binary_service_empty_request = QuadStore()

    def setUp(self) -> None:
        self.anzo_client = AnzoClient(
            SERVER, PORT, username=USERNAME, password=PASSWORD)

    def test_simple_semantic_service_endpoint_call(self) -> None:
        result = self.anzo_client.execute_semantic_service(
            self.reload_graphmart_service_uri,
            self.reload_graphmart_request
        )
        named_graph_quads = result.filter(
            graph='http://openanzo.org/semanticServices/'
        )
        self.assertNotEqual(named_graph_quads.quads, set())

    def test_simple_service_call_with_empty_request(self) -> None:
        self.anzo_client.execute_semantic_service(
            self.binary_service_uri,
            self.binary_service_empty_request
        )

    def test_bad_semantic_service_uri(self) -> None:
        bad_service_uri = self.reload_graphmart_service_uri = "this_is_bad"

        self.assertRaises(
            RuntimeError, self.anzo_client.execute_semantic_service,
            bad_service_uri, self.reload_graphmart_request
        )

    def test_empty_semantic_service_uri(self) -> None:
        empty_service_uri = ""

        self.assertRaises(
            RuntimeError, self.anzo_client.execute_semantic_service,
            empty_service_uri, self.reload_graphmart_request
        )

    # TODO:
    def test_unsuccessful_semantic_service_call(self) -> None:
        # (where everything used is legit) -> just the service call _fails_
        # e.g., reloading the gmart fails.
        pass



# /*******************************************************************************
#  * Copyright (c) 2019 - 2022 Cambridge Semantics Incorporated.
#  * All rights reserved.
#  * 
#  * Contributors:
#  *     Cambridge Semantics Incorporated
#  *******************************************************************************/

from typing import List
import logging
import time

from .anzo_client import AnzoClient
from .quad_store import QuadStore, Quad
from .query_result import QueryResult
from .uris import TYPE_PRED, ALL_NAMED_GRAPHS_URI


logger = logging.getLogger(__name__)


class GraphmartManager:
    """Class for managing the state of an Anzo Graphmart

    Attributes:
        anzo_client: An AnzoClient object pointing to the Anzo
            with the Graphmart
        graphmart_uri: URI of the Graphmart to manage

    Usage:
        from pyanzo import AnzoClient, GraphmartManager
        anzo_client = AnzoClient("localhost", "8443", "
        graphmart_uri = "http://cambridgesemantics.com/graphmart/123"
        graphmart_manager = GraphmartManager(anzo_client, graphmart_uri)
        graphmart_manager.activate()
    """

    ACTIVATION_URI = "http://cambridgesemantics.com/semanticServices/gqe#deployGraphmart"  # noqa
    DYNAMIC_ACTIVATION_URI = "http://cambridgesemantics.com/semanticServices/Cloud/SpinupWorkflowService#startWorkflow" # noqa
    REFRESH_URI = "http://cambridgesemantics.com/semanticServices/gqe#refreshGraphmart"  # noqa
    RELOAD_URI = "http://cambridgesemantics.com/semanticServices/gqe#reloadGraphmart"  # noqa
    START_WORKFLOW_URI = "http://cambridgesemantics.com/semanticServices/Cloud/TeardownWorkflowService#startWorkflow"  # noqa
    SUCCESS_PREDICATE = "http://cambridgesemantics.com/ontologies/Graphmarts#success"  # noqa

    def __init__(self, anzo_client: AnzoClient, graphmart_uri: str):
        """Constructs a GraphmartManager for managing the state of a Graphmart

        Args:
            anzo_client: An AnzoClient object used for interacting with an Anzo
                server.
            graphmart_uri: The URI of the graphmart. It's assumed that the
                Graphmart is on the server underlying the anzo client object.

        Returns:
            A GraphmartManager object
        """
        self.anzo_client = anzo_client
        self.graphmart_uri = graphmart_uri
        self.sleep_time = 1

    def query(self, query_string: str = "", query_file: str = "",
              data_layers: List[str] = None) -> QueryResult:
        """Queries the Graphmart

        Either query_string or query_file must be provided but not both.

        If no data layers are provided, then all data layers are queried.
        """

        return self.anzo_client.query_graphmart(
            self.graphmart_uri, query_string=query_string,
            query_file=query_file, data_layers=data_layers
        )

    def deactivate(self) -> str:
        """Deactivates the Graphmart

        It is recommended to confirm that the Graphmart is offline by calling
        GraphmartManager.is_graphmart_online after this method completes.
        """

        request = self._build_deactivate_request()

        # Don't check the response because it doesn't contain information.
        # It's best to check if the deactivation failed by
        # calling self.is_graphmart_online

        response = self.anzo_client.execute_semantic_service(
            self.START_WORKFLOW_URI, request
        )
        # sleep one second to let the deactivate propagate
        time.sleep(self.sleep_time)
        return response.to_trig_string()

    def dynamic_activate(self, azg_launch_config: str = "", azg_version:str = "", azg_size: str = "") -> str:
        """Activates the Graphmart and spins up a k8s configured anzograph cluster 
        based on the parameters provided. All parameters are technically optional, but will most likely fail if not provided.
        
        This will silently fail if the launch config, azg version or azg size specified are not valid values.
        Args:
            azg_launch_config (required): The launch configuration containing the hardware specification of the virtual anzograph cluster to be deployed.
            azg_version (required): The version of the anzograph cluster to be deployed
            azg_size (required): The size of the anzograph node pool

        """
        request = self._build_dynamic_activate_request(azg_launch_config, azg_version, azg_size)

        response = self.anzo_client.execute_semantic_service(
             self.DYNAMIC_ACTIVATION_URI, request
        )
        time.sleep(self.sleep_time)  # sleep to let the activate propagate
        return response.to_trig_string()

        
        #
    def activate(self, azg_uri: str = "") -> str:
        """Activates the Graphmart

        If there are multiple AnzoGraph datasources and an
        AnzoGraph URI is not specified, then an arbitrary AnzoGraph
        datasource will be used (this is Anzo behavior,
        not GraphmartManager behavior).

        Note that layers can fail to load and a runtime error won't
        be raised.

        It is recommended to confirm that the Graphmart is online
        and that the expected layers are online after activating the Graphmart.

        If the activate call fails, a runtime error will be raised.

        In more detail:
        1. The semantic service request will internally fail if the
           graph fails to activate
        2. This results in a 500 error being returned from the AnzoClient
           endpoint
        3. Pyanzo will raise a 500 error

        If the Graphmart is already activated, then an error is _not_ thrown.

        Args:
            azg_uri (optional): URI of the AnzoGraph datasource.
        """

        request = self._build_activate_request(azg_uri)

        # Don't check the response because it doesn't contain information. It's
        # best to check if the activation failed by (a) the method
        # execute_semantic_service didn't raise an error and (b) by checking
        # that the Graphmart and expected layers are online

        response = self.anzo_client.execute_semantic_service(
            self.ACTIVATION_URI, request
        )
        time.sleep(self.sleep_time)  # sleep to let the activate propagate
        return response.to_trig_string()

    def refresh(self) -> str:
        """Refreshes the Graphmart

        Note that layers can fail to load and a runtime error won't
        be raised.
        """
        request = self._build_refresh_request()
        response = self.anzo_client.execute_semantic_service(
            self.REFRESH_URI, request
        )

        # This will be true even if some layers failed
        success_quadstore = response.filter(
            predicate=self.SUCCESS_PREDICATE, statement_object="true"
        )
        if len(success_quadstore) != 1:
            raise RuntimeError(f"Refreshing Graphmart {self} failed.")
        # sleep to let the refresh propagate
        time.sleep(self.sleep_time)
        try:
            res = response.to_trig_string()
        except:
            res = response
        return res

    def reload(self) -> str:
        """Reloads the Graphmart.

        Note that layers can fail to load and a runtime error won't
        be raised.
        """
        request = self._build_refresh_request()
        response = self.anzo_client.execute_semantic_service(
            self.RELOAD_URI, request
        )
        time.sleep(self.sleep_time)  # sleep to let the reload propagate
        try:
            res = response.to_trig_string()
        except:
            res = response
        return res

    def is_graphmart_online(self) -> bool:
        """Return True if the Graphmart is online and false otherwise.

        Notes
        - If some layers failed to load, this may return True. It's better to
        run are_layers_online
        - If the Graphmart is activating, refreshing, or reloading, it may
        be online, depending on the Graphmart settings
        """

        query = self._build_are_online_query([self.graphmart_uri])
        query_res = self.anzo_client.query_system_tables(query)
        record_dicts = query_res.as_table_results().as_record_dictionaries()

        error_msg = (
            "Query checking the status of the Graphmart"
            "returned ambiguous results"
        )

        if len(record_dicts) != 1:
            raise RuntimeError(error_msg)

        row = record_dicts[0]
        if row["uri"] != self.graphmart_uri:
            raise RuntimeError(error_msg)

        return "System#Online" in row["status"]

    def are_layers_online(self, layer_uris: List[str] = None) -> bool:
        """Returns True if the specified layers are online. If layer_uris is
        None or empty empty then it will return True if all layers are
        online and False if at least one layer is not online

        TODO: mayabe this only check activated layers, if the default is specified.
        """

        layer_uris = layer_uris if layer_uris else self.get_layers()
        if not layer_uris:
            raise RuntimeError(
                "Cannot determine if layers are online because"
                "there are no layers in the Graphmart"
            )

        return self._are_assets_online(layer_uris)

    def _are_assets_online(self, uris: List[str]) -> bool:
        if not uris:
            raise ValueError("List of uris must be non-empty")

        query = self._build_are_online_query(uris)
        query_res = self.anzo_client.query_system_tables(query_string=query)

        are_all_online = True

        for row in query_res.as_table_results().as_record_dictionaries():
            uri = row["uri"]

            is_online = "System#Online" in row["status"]
            if not is_online:
                status = row["status"]
                logger.debug(f"Asset {uri} isn't online; its state: {status}")
                are_all_online = False

        return are_all_online

    def get_layers(self) -> List[str]:
        layer_query = """
            SELECT DISTINCT ?layer WHERE {
                ?gmart <http://openanzo.org/ontologies/2008/07/Anzo#defaultNamedGraph> ?layer .
                ?layer a <http://cambridgesemantics.com/ontologies/Graphmarts#Layer> .
                FILTER(?gmart=<{GRAPHMART}>)
            }
        """.replace("{GRAPHMART}", self.graphmart_uri)  # noqa
        layers_lists = self.anzo_client.query_journal(
            layer_query, named_graphs=ALL_NAMED_GRAPHS_URI
        ).as_table_results().as_list()

        return [l for layer_list in layers_lists for l in layer_list]

    def get_steps(self) -> List[str]:
        step_query = """
            SELECT DISTINCT ?step WHERE {
                ?gmart <http://openanzo.org/ontologies/2008/07/Anzo#defaultNamedGraph> ?step .
                ?step a <http://cambridgesemantics.com/ontologies/Graphmarts#Step> .
                FILTER(?gmart=<{GRAPHMART}>)
            }
        """.replace("{GRAPHMART}", self.graphmart_uri)  # noqa
        step_lists = self.anzo_client.query_journal(
            step_query, named_graphs=ALL_NAMED_GRAPHS_URI
        ).as_table_results().as_list()

        return [s for step_list in step_lists for s in step_list]

    def enable_layers(self, layer_uris: List[str] = None) -> None:
        """Enables the list of layers provided. If layer uris is none, then
        all layers are enabled.
        """

        layer_uris_list = layer_uris if layer_uris else []
        query = self._build_enable_disable_query(True, layer_uris_list)
        self.anzo_client.update_journal(query)
        time.sleep(self.sleep_time)  # sleep to let the update propagate

    def disable_layers(self, layer_uris: List[str] = None) -> None:
        """Disables the list of layers provided. If layer uris is none, then
        all layers are disabled
        """

        layer_uris_list = layer_uris if layer_uris else []
        query = self._build_enable_disable_query(False, layer_uris_list)
        self.anzo_client.update_journal(query)
        time.sleep(self.sleep_time)  # sleep to let the update propagate

    def enable_steps(self, step_uris: List[str] = None) -> None:
        """Enables the list of steps provided. If steps uris is none, then
        all steps are enabled.
        """

        step_uris_list = step_uris if step_uris else []
        query = self._build_enable_disable_query(True, step_uris_list)
        self.anzo_client.update_journal(query)
        time.sleep(self.sleep_time)  # sleep to let the update propagate

    def disable_steps(self, step_uris: List[str] = None) -> None:
        """Disables the list of steps provided. If steps uris is none, then
        all steps are enabled.
        """

        step_uris_list = step_uris if step_uris else []
        query = self._build_enable_disable_query(False, step_uris_list)
        self.anzo_client.update_journal(query)
        time.sleep(self.sleep_time)  # sleep to let the update propagate

    def are_steps_online(self, step_uris: List[str] = None) -> None:
        """Returns True if the specified steps are online
        """

        step_uris = step_uris if step_uris else self.get_steps()
        if not step_uris:
            raise RuntimeError(
                "Cannot determine if steps are online because"
                "there are no steps in the Graphmart"
            )

        return self._are_assets_online(step_uris)

    def _make_values_clause(self, var_name: str, values: List[str]) -> str:
        if not values:
            raise ValueError("Values list must be non-empty")

        encoded_values = [f"({val})" for val in values]

        ret = f" VALUES ({var_name}) {{ "
        ret += " ".join(encoded_values)
        ret += " } "
        return ret

    def _build_are_online_query(self, uris: List[str]) -> str:
        query_base = """
        SELECT DISTINCT ?uri ?status
        WHERE {
            OPTIONAL { ?uri <http://cambridgesemantics.com/ontologies/GraphmartStatus#status> ?status . }
            {VALUES}
        }
        """  # noqa

        if not uris:
            msg = "At least one URI must be provided for building query"
            raise ValueError(msg)

        uris_with_brackers = [f"<{uri}>" for uri in uris]
        values_sub = self._make_values_clause(
            "?uri", uris_with_brackers
        )
        query = query_base.replace("{VALUES}", values_sub)
        return query

    def _build_enable_disable_query(self, should_enable: bool,
                                    uris: List[str]) -> str:
        query_base = """
        DELETE {
            GRAPH ?uri {
                ?uri <http://cambridgesemantics.com/ontologies/Graphmarts#enabled> {OLD_STATE} .
            }
        }
        INSERT {
            GRAPH ?uri {
                ?uri <http://cambridgesemantics.com/ontologies/Graphmarts#enabled> {NEW_STATE} .
            }
        }

        WHERE {
            { SELECT DISTINCT ?uri WHERE {
                GRAPH ?gmart {
                    ?gmart <http://openanzo.org/ontologies/2008/07/Anzo#defaultNamedGraph> ?uri .
                }
            }}

            GRAPH ?uri {
                ?uri <http://cambridgesemantics.com/ontologies/Graphmarts#enabled> {OLD_STATE} .
            }

            {VALUES_GMART}
            {VALUES_URIS}
        }
        """  # noqa

        if should_enable:
            old_state_sub = "false"
            new_state_sub = "true"
        else:
            old_state_sub = "true"
            new_state_sub = "false"

        if uris:
            encoded_uris = [f"(<{uri}>)" for uri in uris]
            layers_sub = "VALUES (?uri) { "
            layers_sub += " ".join(encoded_uris)
            layers_sub += " } "
        else:
            layers_sub = ""

        if uris:
            uri_values = self._make_values_clause("?uri", uris)
        else:
            uri_values = ""

        gmart_sub = f"VALUES (?gmart) {{ ( <{self.graphmart_uri}> ) }}"

        query = query_base.replace("{VALUES_GMART}", gmart_sub)
        query = query.replace("{VALUES_URIS}", layers_sub)
        query = query.replace("{OLD_STATE}", old_state_sub)
        query = query.replace("{NEW_STATE}", new_state_sub)
        return query

    def _build_activate_request(self, azg_uri: str = "") -> QuadStore:
        sub = "http://openanzo.org/semanticServices#graphmartRequest"
        graph = "http://openanzo.org/semanticServices#graphmartRequest"

        type_obj = "http://cambridgesemantics.com/ontologies/Graphmarts#DeployGraphmartRequest"  # noqa
        type_quad = Quad(sub, TYPE_PRED, type_obj, graph)

        block_pred = "http://cambridgesemantics.com/ontologies/Graphmarts#blockTillCreated"  # noqa
        block_obj = "true"
        block_quad = Quad(sub, block_pred, block_obj, graph, obj_type="literal")  # noqa

        gmart_pred = "http://cambridgesemantics.com/ontologies/Graphmarts#graphmart"  # noqa
        gmart_quad = Quad(sub, gmart_pred, self.graphmart_uri, graph)
        quads = {type_quad, block_quad, gmart_quad}

        if azg_uri:
            azg_pred = "http://cambridgesemantics.com/ontologies/Graphmarts#graphQueryEngineUri"  # noqa
            quads.add(Quad(sub, azg_pred, azg_uri, graph))

        return QuadStore(quads)

    def _build_refresh_request(self, layers: List[str] = None) -> QuadStore:
        sub = "http://openanzo.org/semanticServices#graphmartRequest"
        graph = "http://openanzo.org/semanticServices#graphmartRequest"

        type_obj = "http://cambridgesemantics.com/ontologies/Graphmarts#GraphmartRequest"  # noqa
        type_quad = Quad(sub, TYPE_PRED, type_obj, graph)

        gmart_pred = "http://cambridgesemantics.com/ontologies/Graphmarts#graphmart"  # noqa
        gmart_quad = Quad(sub, gmart_pred, self.graphmart_uri, graph)

        quads = {type_quad, gmart_quad}

        layers = layers if layers else []
        for layer in layers:
            layer_pred = "http://cambridgesemantics.com/ontologies/Graphmarts#graphmartChildUri"  # noqa
            layer_quad = Quad(sub, layer_pred, layer, graph)
            quads.add(layer_quad)

        return QuadStore(quads)

    def _build_deactivate_request(self) -> QuadStore:
        sub = "http://openanzo.org/semanticServices#graphmartRequest"
        graph = "http://openanzo.org/semanticServices#graphmartRequest"

        type_objs = [
            "http://cambridgesemantics.com/ontologies/cloud/workflow/WorkflowRequest",  # noqa
            "http://cambridgesemantics.com/ontologies/cloud/workflow/graphmart/WorkflowGMRequest",  # noqa
            "http://openanzo.org/ontologies/2008/2007/SemanticService#ServiceRequest"  # noqa
        ]

        gmart_pred = "http://cambridgesemantics.com/ontologies/cloud/workflow/graphmart/graphmart"  # noqa

        quads = {
            Quad(sub, TYPE_PRED, type_obj, graph) for type_obj in type_objs
        }

        gmart_quad = Quad(sub, gmart_pred, self.graphmart_uri, graph)
        quads.add(gmart_quad)
        return QuadStore(quads)

    def _build_dynamic_activate_request(self, azg_launch_config: str = "", azg_version:str = "", azg_size: str = "") -> QuadStore:
        sub = "http://openanzo.org/semanticServices#graphmartRequest"
        graph = "http://openanzo.org/semanticServices#graphmartRequest"

        type_objs = [
            "http://http://cambridgesemantics.com/ontologies/cloud/workflow/WorkflowCloudDeployAZG",  # noqa
            "http://cambridgesemantics.com/ontologies/cloud/workflow/WorkflowRequest",  # noqa
            "http://cambridgesemantics.com/ontologies/cloud/workflow/graphmart/WorkflowGMRequest"  # noqa
        ]
        
        gmart_pred = "http://cambridgesemantics.com/ontologies/cloud/workflow/graphmart/graphmart"  # noqa
        gmart_quad = Quad(sub, gmart_pred, self.graphmart_uri, graph)
        quads = {
            Quad(sub, TYPE_PRED, type_obj, graph) for type_obj in type_objs
        }
        quads.add(gmart_quad)

        selected_azg_pred = "http://cambridgesemantics.com/ontologies/cloud/workflow/selectedAZG"
        
        # As of Anzo 5.3.10, this (selected_azg_pred) is a required predicate.
        quads.add(Quad(sub, selected_azg_pred, "http://moreinfo/spinUp", graph))
        
        if azg_launch_config:
            launch_config_pred = "http://cambridgesemantics.com/ontologies/cloud/workflow/selectedAZGLaunchConfiguration" #noqa
            launch_config_quad = Quad(sub, launch_config_pred, azg_launch_config, graph)
            quads.add(launch_config_quad)
            
        if azg_version:
            version_pred = "http://cambridgesemantics.com/ontologies/cloud/workflow/selectedAZGVersion"  # noqa
            version_quad = Quad(sub, version_pred, azg_version, graph, obj_type="literal")
            quads.add(version_quad)
            
        if azg_size:
            cluster_size_pred = "http://cambridgesemantics.com/ontologies/cloud/workflow/selectedAZGClusterSize"  # noqa
            cluster_size_quad = Quad(sub, cluster_size_pred, azg_size, graph, obj_type="literal", obj_data_type="http://www.w3.org/2001/XMLSchema#int")
            quads.add(cluster_size_quad)
            
        
        return QuadStore(quads)



# /*******************************************************************************
#  * Copyright (c) 2019 - 2022 Cambridge Semantics Incorporated.
#  * All rights reserved.
#  * 
#  * Contributors:
#  *     Cambridge Semantics Incorporated
#  *******************************************************************************/

import copy
import os
import rdflib
import rdflib.compare
import unittest
from pyanzo import AnzoClient, Quad, QuadStore

from .test_common import (
    SERVER,
    PORT,
    USERNAME,
    PASSWORD,
)


class TestQuadStore(unittest.TestCase):
    tmp_filename = "tests/test_assets/tmp.trig"

    trig_filename = "tests/test_assets/sample.trig"
    n3_filename = "tests/test_assets/sample.n3"
    ttl_filename = "tests/test_assets/sample.ttl"
    trix_filename = "tests/test_assets/sample.trix"
    json_ld_filename = "tests/test_assets/sample.json"
    trig_with_bnodes_filename = "tests/test_assets/trig_with_bnodes.trig"

    rdflib_graph = rdflib.ConjunctiveGraph()
    expected_num_statements = 3

    record_dicts = [
        {
            's': 'http://csi.test.com/subject1',
            'p': 'http://csi.test.com/predicate1',
            'o': 'http://csi.test.com/object1',
            'g': 'http://csi.test.com/'
        }, {
            's': 'http://csi.test.com/subject2',
            'p': 'http://csi.test.com/predicate2',
            'o': 'http://csi.test.com/object2',
            'g': 'http://csi.test.com/'
        }, {
            's': 'http://csi.test.com/subject2',
            'p': 'http://csi.test.com/predicate3',
            'o': 7,
            'g': 'http://csi.test.com/graph2'
        }
    ]

    record_dicts_without_graphs = [
        {
            's': 'http://csi.test.com/subject1',
            'p': 'http://csi.test.com/predicate1',
            'o': 'http://csi.test.com/object1'
        }, {
            's': 'http://csi.test.com/subject2',
            'p': 'http://csi.test.com/predicate2',
            'o': 'http://csi.test.com/object2'
        }, {
            's': 'http://csi.test.com/subject2',
            'p': 'http://csi.test.com/predicate3',
            'o': 7
        }
    ]

    quad_set = {
        Quad("http://csi.test.com/subject1",
             "http://csi.test.com/predicate1",
             "http://csi.test.com/object1",
             "http://csi.test.com/"),
        Quad("http://csi.test.com/subject2",
             "http://csi.test.com/predicate2",
             "http://csi.test.com/object2",
             "http://csi.test.com/"),
        Quad("http://csi.test.com/subject2",
             "http://csi.test.com/predicate3",
             "7",
             "http://csi.test.com/graph2",
             obj_type="literal",
             obj_data_type="http://www.w3.org/2001/XMLSchema#integer"),
    }

    quad_store = QuadStore(copy.deepcopy(quad_set))

    quad_store_without_graphs = QuadStore(
        {
            Quad(q.sub, q.pred, q.obj, None, q.sub_type,
                 q.obj_type, q.obj_data_type)
            for q in quad_set
        }
    )

    def setUp(self) -> None:
        self.rdflib_graph.parse(self.trig_filename, format="trig")

        self.anzo_client = AnzoClient(
            SERVER, PORT, username=USERNAME, password=PASSWORD
        )

    def test_from_anzo_json(self) -> None:
        json_one = {"namedGraphUri": "http://csi.test.com/",
                    "subject": {
                        "objectType": "uri",
                        "value": "http://csi.test.com/subject1",
                    },
                    "predicate": "http://csi.test.com/predicate1",
                    "object": {
                        "objectType": "uri",
                        "value": "http://csi.test.com/object1"
                    }
                    }
        json_two = {"namedGraphUri": "http://csi.test.com/",
                    "subject": {
                        "objectType": "uri",
                        "value": "http://csi.test.com/subject2",
                    },
                    "predicate": "http://csi.test.com/predicate2",
                    "object": {
                        "objectType": "uri",
                        "value": "http://csi.test.com/object2"
                    }
                    }

        json_three = {"namedGraphUri": "http://csi.test.com/graph2",
                      "subject": {
                          "objectType": "uri",
                          "value": "http://csi.test.com/subject2",
                      },
                      "predicate": "http://csi.test.com/predicate3",
                      "object": {
                          "dataType": "http://www.w3.org/2001/XMLSchema#integer",
                          "objectType": "literal",
                          "value": "7"
                      }
                      }

        quad_store = QuadStore.from_anzo_json_list(
            [json_one, json_two, json_three]
        )
        self.assertEqual(quad_store, self.quad_store)

    def test_from_anzo_json_error(self) -> None:
        json = {"Penny": 1}
        quad_store = QuadStore()
        self.assertRaises(RuntimeError, quad_store.from_anzo_json_list, json)

    def test_from_rdflib_graph(self) -> None:
        qs = QuadStore.from_rdflib_graph(self.rdflib_graph)
        self.assertEqual(qs, self.quad_store)

        # Test with a non-conjuctive graph
        graph = rdflib.ConjunctiveGraph()
        graph.parse(self.trig_filename, format="trig")
        qs = QuadStore.from_rdflib_graph(self.rdflib_graph)
        self.assertEqual(qs, self.quad_store)

    def test_from_trig_file(self) -> None:
        qs = QuadStore.from_trig_file(self.trig_filename)
        self.assertEqual(qs, self.quad_store)

    def test_to_trig_file(self) -> None:
        self.quad_store.write_to_trig_file(self.tmp_filename)
        qs = QuadStore.from_trig_file(self.tmp_filename)
        self.assertEqual(qs, self.quad_store)
        os.remove(self.tmp_filename)

    def test_from_ttl_file(self) -> None:
        qs = QuadStore.from_file(self.ttl_filename, "ttl")
        self.assertEqual(qs, self.quad_store_without_graphs)

    def test_to_ttl_file(self) -> None:
        self.quad_store.write_to_file(self.tmp_filename, "ttl")
        qs = QuadStore.from_file(self.tmp_filename, "ttl")
        self.assertEqual(qs, self.quad_store_without_graphs)
        os.remove(self.tmp_filename)

    def test_from_n3_file(self) -> None:
        qs = QuadStore.from_file(self.n3_filename, "n3")
        self.assertEqual(qs, self.quad_store_without_graphs)

    def test_to_n3_file(self) -> None:
        self.quad_store.write_to_file(self.tmp_filename, "n3")
        qs = QuadStore.from_file(self.tmp_filename, "n3")
        self.assertEqual(qs, self.quad_store_without_graphs)
        os.remove(self.tmp_filename)

    def test_from_trix_file(self) -> None:
        qs = QuadStore.from_file(self.trix_filename, "trix")
        self.assertEqual(qs, self.quad_store_without_graphs)

    def test_to_trix_file(self) -> None:
        self.quad_store.write_to_file(self.tmp_filename, "trix")
        qs = QuadStore.from_file(self.tmp_filename, "trix")
        self.assertEqual(qs, self.quad_store_without_graphs)
        os.remove(self.tmp_filename)

    def test_from_json_ld_file(self) -> None:
        qs = QuadStore.from_file(self.json_ld_filename, "json-ld")
        self.assertEqual(qs, self.quad_store)

    def test_to_json_ld_file(self) -> None:
        self.quad_store.write_to_file(self.tmp_filename, "json-ld")
        qs = QuadStore.from_file(self.tmp_filename, "json-ld")
        self.assertEqual(qs, self.quad_store)
        os.remove(self.tmp_filename)

    def test_as_rdflib_graph(self) -> None:
        rdflib_graph = self.quad_store.as_rdflib_graph()

        self.assertTrue(
            rdflib.compare.isomorphic(rdflib_graph, self.rdflib_graph)
        )

    def test_as_rdflib_graph_with_bnodes(self) -> None:
        sub = "subject2"
        pred = "http://csi.test.com/predicate3"
        obj = "http://csi.test.com/object3"
        graph = "http://csi.test.com/graph2"

        quad_store = QuadStore({Quad(sub, pred, obj, graph, "bnode", "uri")})
        g = quad_store.as_rdflib_graph()

        expected_g = rdflib.ConjunctiveGraph()
        expected_g.add((
            rdflib.BNode(sub),
            rdflib.URIRef(pred),
            rdflib.URIRef(obj),
            rdflib.URIRef(graph),
        ))

        self.assertTrue(
            rdflib.compare.isomorphic(g, expected_g)
        )

    def test_as_record_dictionaries(self) -> None:
        record_dicts = self.quad_store.as_record_dictionaries()

        self.assertEqual(
            sorted(record_dicts,
                   key=lambda d: (d['s'], d['p'], d['o'], d['g'])),
            sorted(self.record_dicts,
                   key=lambda d: (d['s'], d['p'], d['o'], d['g']))
        )

    def test_as_record_dictionaries_without_graphs(self) -> None:
        record_dicts = self.quad_store_without_graphs.as_record_dictionaries()
        self.assertEqual(
            sorted(record_dicts, key=lambda d: (d['s'], d['p'], d['o'])),
            sorted(self.record_dicts_without_graphs,
                   key=lambda d: (d['s'], d['p'], d['o']))
        )

    def test_quad_store_iterator_correct(self) -> None:
        quad_store = QuadStore(copy.deepcopy(self.quad_set))
        result = [quad for quad in quad_store]
        self.assertEqual(sorted(result), sorted(list(self.quad_set)))

    def test_filtering_by_nothing(self) -> None:
        quad_store = QuadStore(copy.deepcopy(self.quad_set))
        filtered_store = quad_store.filter()
        self.assertEqual(quad_store, filtered_store)

    def test_filtering_by_subject(self) -> None:
        quad_store = QuadStore(copy.deepcopy(self.quad_set))
        result_store = quad_store.filter(
            subject="http://csi.test.com/subject2"
        )

        expected_store = QuadStore({
            Quad("http://csi.test.com/subject2",
                 "http://csi.test.com/predicate3",
                 "7",
                 "http://csi.test.com/graph2",
                 obj_type="literal",
                 obj_data_type="http://www.w3.org/2001/XMLSchema#integer"),
            Quad("http://csi.test.com/subject2",
                 "http://csi.test.com/predicate2",
                 "http://csi.test.com/object2",
                 "http://csi.test.com/")
        })

        self.assertEqual(expected_store, result_store)

    def test_add_basic(self) -> None:
        qs = QuadStore()
        for quad in self.quad_set:
            new = qs.add(quad)
            self.assertTrue(new)
        self.assertEqual(self.quad_store, qs)

    def test_add_repeat(self) -> None:
        qs = QuadStore()
        for quad in self.quad_set:
            qs.add(quad)

        quad = next(iter(self.quad_set))
        new = qs.add(quad)
        self.assertEqual(self.quad_store, qs)
        self.assertFalse(new)

    def test_remove(self) -> None:
        qs = QuadStore()
        for quad in self.quad_set:
            qs.add(quad)
        quad = next(iter(self.quad_set))
        removed = qs.remove(quad)
        self.assertEqual(len(qs), len(self.quad_store) - 1)
        self.assertTrue(removed)

    def test_remove_nonexist(self) -> None:
        qs = QuadStore()
        for quad in self.quad_set:
            qs.add(quad)
        quad = next(iter(self.quad_set))
        qs.remove(quad)
        removed = qs.remove(quad)  # this one should be a no-op, return false
        self.assertEqual(len(qs), len(self.quad_store) - 1)
        self.assertFalse(removed)

    def test_set_basic(self) -> None:
        qs = QuadStore()
        for quad in self.quad_set:
            qs.add(quad)

        update_quad = Quad("http://csi.test.com/subject1",
                           "http://csi.test.com/predicate1",
                           "http://csi.test.com/object2",
                           "http://csi.test.com/")

        # should upsert existing statement - object1 changes to object2
        upsert = qs.set(update_quad)

        self.assertTrue(upsert)
        self.assertEqual(len(qs), len(self.quad_store))
        # this should be the updated quad
        quad = qs.filter(subject=update_quad.sub, predicate=update_quad.pred,
                         graph=update_quad.graph).quads.pop()
        self.assertEqual(quad.obj, "http://csi.test.com/object2")

    def test_set_multiple_previous(self) -> None:
        qs = QuadStore()
        for quad in self.quad_set:
            qs.add(quad)

        add_quad = Quad("http://csi.test.com/subject1",
                        "http://csi.test.com/predicate1",
                        "http://csi.test.com/object2",
                        "http://csi.test.com/")
        qs.add(add_quad)  # add a new statement
        # qs should be 1 greater than base quad_store because we added a quad
        self.assertEqual(len(qs), len(self.quad_store) + 1)

        update_quad = Quad("http://csi.test.com/subject1",
                           "http://csi.test.com/predicate1",
                           "http://csi.test.com/object3",
                           "http://csi.test.com/")
        upsert = qs.set(update_quad)

        self.assertTrue(upsert)
        # should now be same size as original because we've reset the statement
        self.assertEqual(len(qs), len(self.quad_store))

        # this should be the updated quad
        quad = qs.filter(subject=update_quad.sub, predicate=update_quad.pred,
                         graph=update_quad.graph).quads.pop()
        self.assertEqual(quad.obj, "http://csi.test.com/object3")

    def test_set_none_previous(self) -> None:
        qs = QuadStore()
        for quad in self.quad_set:
            qs.add(quad)

        update_quad = Quad("http://csi.test.com/subject1",
                           "http://csi.test.com/predicate4",
                           "http://csi.test.com/object4",
                           "http://csi.test.com/")
        # should insert a new statement
        upsert = qs.set(update_quad)

        self.assertFalse(upsert)
        self.assertEqual(len(qs), len(self.quad_store) + 1)
        # this should be the new quad
        quad = qs.filter(subject=update_quad.sub, predicate=update_quad.pred,
                         graph=update_quad.graph).quads.pop()
        self.assertEqual(quad.obj, "http://csi.test.com/object4")

    def test_trig_with_bnodes(self):
        qs = QuadStore.from_trig_file(self.trig_with_bnodes_filename)
        g = qs.as_rdflib_graph()

        rdflib_graph = rdflib.ConjunctiveGraph()
        rdflib_graph.parse(self.trig_with_bnodes_filename, format="trig")
        self.assertTrue(
            rdflib.compare.isomorphic(rdflib_graph, qs.as_rdflib_graph())
        )

    def test_to_trig_string(self) -> None:
        s = self.quad_store.to_trig_string()
        qs = QuadStore.from_trig_string(s)
        self.assertEqual(qs, self.quad_store)

    def test_to_trig_string2(self) -> None:
        s = self.quad_store.to_string("trig")
        qs = QuadStore.from_string(s, "trig")
        self.assertEqual(qs, self.quad_store)

    def test_to_from_json_ld_string(self) -> None:
        s = self.quad_store.to_string("json-ld")
        qs = QuadStore.from_string(s, "json-ld")
        self.assertEqual(qs, self.quad_store)

    def test_to_from_ttl_string(self) -> None:
        s = self.quad_store.to_string("ttl")
        qs = QuadStore.from_string(s, "ttl")
        self.assertEqual(qs, self.quad_store_without_graphs)

    def test_to_from_n3_string(self) -> None:
        s = self.quad_store.to_string("n3")
        qs = QuadStore.from_string(s, "n3")
        self.assertEqual(qs, self.quad_store_without_graphs)

    def test_to_from_trix_string(self) -> None:
        s = self.quad_store.to_string("trix")
        qs = QuadStore.from_string(s, "trix")
        self.assertEqual(qs, self.quad_store_without_graphs)

    def test_to_string_bad_format(self) -> None:
        self.assertRaises(
            Exception, self.quad_store.to_string, "foo"
        )

    def test_form_string_bad_format(self) -> None:
        s = self.quad_store.to_string("trig")

        self.assertRaises(
            Exception, QuadStore.from_string, s, "ttl"
        )

        self.assertRaises(
            Exception, QuadStore.from_string, s, "foo"
        )

        self.assertRaises(
            Exception, QuadStore.from_string, s, ""
        )



import unittest
from pyanzo import AnzoClient, QuadStore, Quad

from .test_common import (
    SERVER,
    PORT,
    USERNAME,
    PASSWORD
)


class TestAnzoClientUpdateJournal(unittest.TestCase):
    graph_uri = "urn://graph1"

    wipe_graph_query = """
        DELETE { GRAPH ?g { ?s ?p ?o } }
        WHERE { GRAPH ?g { ?s ?p ?o } FILTER(?g=<{GRAPH}>) }
    """.replace("{GRAPH}", graph_uri)


    seed_graph_query = """
        PREFIX dc: <http://purl.org/dc/elements/1.1/>

        INSERT DATA {
            GRAPH <{GRAPH}> {
                <urn://sub> <urn://pred> <urn://obj1> .
                <urn://sub> <urn://pred> <urn://obj2> .
            }
        }
    """.replace("{GRAPH}", graph_uri)  # noqa

    insert_delete_query_with_graph = """
        DELETE {
            graph ?g {
                <urn://sub> <urn://pred> ?o .
            }
        }
        INSERT {
            graph ?g {
                <urn://sub> <urn://pred> <urn://obj3>
            }
        }
        WHERE {
            graph ?g {
                <urn://sub> <urn://pred> ?o .
                FILTER(?o = <urn://obj2>)
            }
        }
    """

    insert_delete_query_without_graph = """
        DELETE {
            graph <{GRAPH}> {
                <urn://sub> <urn://pred> ?o .
            }
        }
        INSERT {
            graph <{GRAPH}> {
                <urn://sub> <urn://pred> <urn://obj3>
            }
        }
        WHERE {
            <urn://sub> <urn://pred> ?o .
            FILTER(?o = <urn://obj2>)
        }
    """.replace("{GRAPH}", graph_uri)

    insert_without_insert_graph = """
        DELETE {
            graph <{GRAPH}> {
                <urn://sub> <urn://pred> ?o .
            }
        }
        INSERT {
            <urn://sub> <urn://pred> <urn://obj3>
        }
        WHERE {
            <urn://sub> <urn://pred> ?o .
            FILTER(?o = <urn://obj2>)
        }
    """.replace("{GRAPH}", graph_uri)

    insert_without_delete_graph = """
        DELETE {
            <urn://sub> <urn://pred> ?o .
        }
        INSERT {
            graph <{GRAPH}> {
                <urn://sub> <urn://pred> <urn://obj3>
            }
        }
        WHERE {
            <urn://sub> <urn://pred> ?o .
            FILTER(?o = <urn://obj2>)
        }
    """.replace("{GRAPH}", graph_uri)

    quad1 = Quad("urn://sub", "urn://pred", "urn://obj1", graph_uri)
    quad2 = Quad("urn://sub", "urn://pred", "urn://obj2", graph_uri)
    quad3 = Quad("urn://sub", "urn://pred", "urn://obj3", graph_uri)

    quadstore12 = QuadStore({quad1, quad2})  # contains quads 1 and 2
    quadstore13 = QuadStore({quad1, quad3})  # contains quads 1 and 3

    def setUp(self) -> None:
        self.anzo_client = AnzoClient(
            server=SERVER, username=USERNAME,
            password=PASSWORD, port=PORT
        )

    def wipe_graph(self) -> None:
        self.anzo_client.update_journal(self.wipe_graph_query)

    def seed_graph(self) -> None:
        self.wipe_graph()
        self.anzo_client.update_journal(self.seed_graph_query)

        graph = self.anzo_client.get(self.graph_uri).filter(
            graph=self.graph_uri
        )
        self.assertEqual(graph, self.quadstore12)

    def test_insert_data(self) -> None:
        self.wipe_graph()
        self.anzo_client.update_journal(self.seed_graph_query)
        quadstore = self.anzo_client.get(self.graph_uri).filter(
            graph=self.graph_uri
        )
        self.assertEqual(quadstore, self.quadstore12)

    def test_insert_delete_with_graph(self) -> None:
        self.seed_graph()
        self.anzo_client.update_journal(self.insert_delete_query_with_graph)
        graph = self.anzo_client.get(self.graph_uri).filter(
            graph=self.graph_uri
        )
        self.assertEqual(graph, self.quadstore13)

    def test_insert_delete_without_graph(self) -> None:
        self.seed_graph()
        self.anzo_client.update_journal(self.insert_delete_query_without_graph)
        graph = self.anzo_client.get(self.graph_uri).filter(
            graph=self.graph_uri
            )
        self.assertEqual(graph, self.quadstore13)

    def test_insert_without_insert_graph(self) -> None:
        self.seed_graph()
        self.assertRaises(
            RuntimeError,
            self.anzo_client.update_journal,
            self.insert_without_insert_graph
        )

    def test_insert_without_delete_graph(self) -> None:
        self.seed_graph()
        self.assertRaises(
            RuntimeError,
            self.anzo_client.update_journal,
            self.insert_without_delete_graph
        )




# /*******************************************************************************
#  * Copyright (c) 2019 - 2022 Cambridge Semantics Incorporated.
#  * All rights reserved.
#  * 
#  * Contributors:
#  *     Cambridge Semantics Incorporated
#  *******************************************************************************/

import json
from typing import List, Iterator
from .quad_store import QuadStore, semantic_type_to_python_object


class TableResult:
    """Wraps the results from certain SPARQL queries and provides useful
    access methods.

    Attributes:
        dict_results: The exact results of the query parsed into a dictionary
            from the response json
        result_vars: The names of all variables returned from the query
        result_dicts: The variable bindings returned from the query

    Usage:
        # TableResults are created from QueryResults
        table_results = query_result.as_table_results()

        # There are multiple different ways to access the results of the table
        table_results.as_list()
        table_results.as_list_iter()
        table_results.as_as_record_dictionaries()
    """

    def __init__(self, json_string: str) -> None:
        self.dict_results = json.loads(json_string)
        self.result_vars = self.dict_results['head']['vars']
        self.result_dicts = self.dict_results['results']['bindings']

    def as_list(self) -> List[List[str]]:
        """
        Returns the query result as a list of lists.

        Convert a query result to a list of lists where each sublist
        represents a single solution to the query. If any variables are unbound
        (e.g. due to the use of OPTIONAL {...}), the corresponding value in the
        list will be an empty string. This means that all sublists will be of
        the same length.

        Returns:
            A list of lists of string values.
        """
        return list(self.as_list_iter())

    def as_list_iter(self) -> Iterator[List[str]]:
        """
        Returns the query result as an iterator containing lists.

        Converts a query result to an iterator containing lists where each
        sublist represents a single solution to the query. If any variables are
        unbound (e.g. due to the use of OPTIONAL {...}), the corresponding
        value in the list will be an empty string. This means that all sublists
        will be of the same length.

        Yields:
            A list string values representing one solution to the query

        """
        for result_dict in self.result_dicts:
            yield [
                result_dict.get(var, {'value': ''})['value']
                for var in self.result_vars
            ]

    def as_record_dictionaries(self) -> List[dict]:
        """
        Returns the query result as list of dictionaries.

        Convert a query result to a list of dictionaries where each dictionary
        represents a single solution to the query and maps each variable to its
        corresponding value within the solution. Any variables that are unbound
        (e.g. due to the use of OPTIONAL {...}) will still appear as keys in
        the dictionary but will map to the empty string. This means that all
        dictionaries will have the same number of keys.

        Returns:
            A list of dictionaries, each representing one solution to the
            query. The return format is perfectly suited to interface with
            pandas through the DataFrame, from_dict or from_records methods.
        """

        record_dicts = []
        for result_dict in self.result_dicts:
            record_dict = {}

            for var_name in self.result_vars:

                # Parse the information in the result dict, if this variable
                # exists. If it doesn't, then use these defaults
                val = ""
                val_datatype = ""

                if var_name in result_dict:
                    # result_dict[var_name] might look something like this:
                    # {'datatype': 'http://www.w3.org/2001/XMLSchema#int',
                    # 'type': 'literal', 'value': '30'}}

                    val = result_dict[var_name]['value']

                    # A datatype is not always provided, so default to
                    # the empty string
                    val_datatype = result_dict[var_name].get("datatype", "")

                # Convert the value to the python type. For example,
                # val might be "3" and the val_datatype might be xsd:int.
                # Then semantic_type_to_python_object(val, val_datatype)
                # would return the int 3.

                val_typed = semantic_type_to_python_object(val, val_datatype)
                record_dict[var_name] = val_typed

            record_dicts.append(record_dict)
        return record_dicts

    def __iter__(self):
        return self.as_list_iter()

    def __eq__(self, other) -> bool:
        return self.dict_results == other.dict_results

    def __str__(self) -> str:
        return f"<TableResult '{self.dict_results}'>"

    def __repr__(self) -> str:
        return str(self)


class QueryResult:
    """Dispatches a query result to a TableResult or QuadStore.

    This class handles the results of queries. And dispatches them
    to either a TableResult class or a QuadStore class. The former
    is for select and ask queries. The latter is for construct queries.

    Attributes:
        json_string: A json string that represents the results
            of a query.
    """

    def __init__(self, json_string: str) -> None:
        """Stores json_string from a query result.

        This method just stashes the json_string it is
        passed as an attibute.

        Args:
            json_string: A json_string representing the reuslts
                of a query
        """
        self.json_string = json_string

    def as_table_results(self) -> TableResult:
        """Returns the query result as a TableResult.

        """

        return TableResult(self.json_string)

    def as_quad_store(self) -> QuadStore:
        """Returns the query result as a QuadStore.

        """

        # Cover for an Anzo issue where an empty string,
        # not an empty JSON, is returned.
        if not self.json_string:
            return QuadStore()

        result_json = json.loads(self.json_string)
        return QuadStore.from_anzo_json_list(result_json)



# /*******************************************************************************
#  * Copyright (c) 2019 - 2022 Cambridge Semantics Incorporated.
#  * All rights reserved.
#  * 
#  * Contributors:
#  *     Cambridge Semantics Incorporated
#  *******************************************************************************/

import unittest

from pyanzo import (
    AnzoClient,
)

from .test_common import (
    SERVER,
    PORT,
    USERNAME,
    PASSWORD,
    GRAPHMART
)


class TestAnzoClientQuerySystemTables(unittest.TestCase):

    query_string = f"""
        PREFIX gm: <http://cambridgesemantics.com/ontologies/GraphmartStatus#>

        SELECT distinct * WHERE {{
            ?graphmart a gm:GraphmartStatus ;
                gm:totalStatements ?statements ;
            .
            FILTER(?graphmart = <{GRAPHMART}>)
        }}
        LIMIT 10
    """  # noqa


    construct_query = """
        PREFIX dc: <http://purl.org/dc/elements/1.1/>
        PREFIX system: <http://openanzo.org/ontologies/2008/07/System#>

        CONSTRUCT {
            <urn://one> dc:title ?type
        } WHERE {
           ?s a ?type .
           #FILTER(?type = system:QueryExecution)
        }
        LIMIT 10
    """  # noqa

    def setUp(self) -> None:
        self.anzo_client = AnzoClient(
            server=SERVER, username=USERNAME,
            password=PASSWORD, port=PORT
        )

    def test_select(self) -> None:
        result = self.anzo_client.query_system_tables(
            query_string=self.query_string
        )

        result_table = result.as_table_results().as_list()

        self.assertEqual(len(result_table), 1)

        only_result = result_table[0]

        self.assertEqual(only_result[0], GRAPHMART)
        self.assertTrue(0 < int(only_result[1]))

    @unittest.skip("Requires Anzo improvement to pass")
    def test_construct(self) -> None:
        result = self.anzo_client.query_system_tables(
            query_string=self.construct_query
        )

        qs = result.as_quad_store()
        self.assertTrue(len(qs) == 10)



# /*******************************************************************************
#  * Copyright (c) 2019 - 2022 Cambridge Semantics Incorporated.
#  * All rights reserved.
#  * 
#  * Contributors:
#  *     Cambridge Semantics Incorporated
#  *******************************************************************************/

import logging
import time
import uuid
from typing import List
import pandas as pd
import urllib.parse

from .anzo_client import AnzoClient
from .quad_store import Quad, QuadStore
from .uris import (
    NAMESPACE,
    TYPE_PRED,
    DNG_PRED,
    NG_TYPE
)

logger = logging.getLogger(__name__)


class LDSManager:
    """Class for managing the state of an Anzo LinkedDataSet
    (i.e "local volume" dataset, "journal" dataset)

    The class can be used to perform operations like
    creating and deleting an Anzo LDS, as well as adding and
    clearing data from an LDS.

    All operations within this class assume that the LDS graph is itself also
        the "Dataset" graph. If this is not the case, then do not use this
        class for LDS management.

    Attributes:
        anzo_client: An AnzoClient object pointing to the Anzo
            with the LDS
        lds_uri: URI of the LDS to manage

    Usage:
        from pyanzo import AnzoClient, LDSManager
        anzo_client = AnzoClient("localhost", "8443")
        lds_uri = "http://cambridgesemantics.com/dataset/123"
        lds_manager = LDSManager(anzo_client, lds_uri)

        from pyanzo import AnzoClient, LDSManager
        anzo_client = AnzoClient("localhost", "8443")
        lds_title = "my test LDS"
        lds_manager = LDSManager.create_lds(anzo_client=anzo_client,
                                            title=lds_title)
    """

    SLEEP_TIME = 3  # need a long sleep time for dataset operations

    def __init__(self, anzo_client: AnzoClient, lds_uri: str):
        """Constructs an LDSManager for managing the state of
        an Anzo LinkedDataSet.

        As with other operations within this class, it is assumed the LDS graph
            is itself also the "Dataset" graph.

        Args:
            anzo_client: An AnzoClient object used for interacting with an Anzo
                server.
            lds_uri: The URI of the LDS. It's assumed that the
                LDS is on the server underlying the anzo client object.

        Returns:
            An LDSManager object
        """
        self.anzo_client = anzo_client
        self.lds_uri = lds_uri

    @classmethod
    def create_lds(cls, anzo_client: AnzoClient, title: str,
                   statements: QuadStore = None,
                   ontologies: List[str] = None) -> 'LDSManager':
        """Creates an LDS on Anzo with the given title.
        Adds the statements in the given quad store to the LDS, if applicable.
        Adds references from the LDS to the given ontologies, if applicable.

        As is assumed with other operations within this class, the LDS graph
            is itself also the "Dataset" graph.

        Args:
            anzo_client: An AnzoClient object used for interacting with an Anzo
                server.
            title: Title to give the LDS that is created.
            statements: (optional) QuadStore of statements to add to the LDS
                once it's creatd
            ontologies: (optional) Ontology URIs that the LDS will reference
                when created via the ld:ontology predicate.


        Returns:
            LDSManager that manages the created LDS
        """

        lds_uri = NAMESPACE["lds"] + str(uuid.uuid1())
        cls._add_lds_graph(anzo_client, lds_uri,
                           title, ontologies)
        cls._register_lds_graph(anzo_client, lds_uri)
        lds_mananger = cls(anzo_client, lds_uri)
        if statements:
            lds_mananger.add_to_lds(statements)
        time.sleep(cls.SLEEP_TIME)
        return lds_mananger

    def add_to_lds(self, statements: QuadStore):
        """Adds the statements in the provided QuadStore to the managed LDS.
        This operation takes care of all that's needed to properly add the
        statements to the LDS; specifically, it adds anzo:defaultNamedGraph
        references from the LDS to the named graphs of the statements

        Args:
            statements: QuadStore of statements to add to the LDS.
                The statements must be quads (they must contain a NamedGraph).

        """
        qs = QuadStore()
        df = pd.DataFrame.from_dict(statements.as_record_dictionaries())
        for g in df['g'].unique():
            qs.add(Quad(self.lds_uri, DNG_PRED, g, self.lds_uri))
        for stmt in statements:
            qs.add(stmt)
        self.anzo_client.add(qs)

    def overwrite_lds(self, statements: QuadStore):
        """Clears the managed LDS of all existing named graphs and statements,
        and then adds the statements in the provided QuadStore to the LDS.
        This is simply a utility method that first executes clear_lds(),
        and then add_to_lds(QuadStore).

        Args:
            statements: QuadStore of statements to add to the LDS.
                The statements must be quads (they must contain a NamedGraph).
        """
        self.clear_lds()
        self.add_to_lds(statements)

    def clear_lds(self):
        """Removes all of the managed LDS's named graphs and their contents
        from the journal. This leaves you with an "empty" LDS - the LDS graph
        itself remains, but it points to nothing.
        """

        # populate quad store with magic triples from all LDS graphs
        to_remove = QuadStore()
        res = self.anzo_client.query_journal("select distinct ?g where { <%s> <%s> ?g . }" % (self.lds_uri, DNG_PRED)).as_table_results().as_list()
        for r in res:
            graph_uri = r.pop()
            # in case LDS references itself, do not delete
            if r != self.lds_uri:
                to_remove.add(Quad(self.lds_uri, DNG_PRED,
                                   graph_uri, self.lds_uri))
                to_remove.add(magic_triple(graph_uri))

        # remove magic triples
        if len(to_remove) > 0:
            self.anzo_client.remove(to_remove)

    def delete_lds(self):
        """Clears the managed LDS of its contents (if applicable) and then
        deletes/deregisters the LDS itself. This method first performs a
        clear_lds() operation, then performs the necessary LDS deregistration.
        """

        self.clear_lds()
        self._deregister_lds_graph(self.anzo_client, self.lds_uri)
        self.anzo_client.remove(QuadStore({magic_triple(self.lds_uri)}))

    @classmethod
    def _add_lds_graph(cls, anzo_client: AnzoClient, lds_uri: str,
                       title: str, ontologies: List[str] = None):

        LDS_TYPE = NAMESPACE["cs-ld"] + "LinkedDataSet"
        DS_TYPE = NAMESPACE["cs-ld"] + "Dataset"
        DS_PRED = NAMESPACE["cs-ld"] + "dataset"
        ONTO_PRED = NAMESPACE["cs-ld"] + "ontology"
        DC_TITLE = NAMESPACE["dc"] + "title"

        qs = QuadStore()
        qs.add(Quad(lds_uri, TYPE_PRED, LDS_TYPE, lds_uri))
        qs.add(Quad(lds_uri, TYPE_PRED, DS_TYPE, lds_uri))
        qs.add(Quad(lds_uri, DC_TITLE, title, lds_uri, obj_type="literal"))
        qs.add(Quad(lds_uri, DS_PRED, lds_uri, lds_uri))

        if ontologies:
            for onto in ontologies:
                qs.add(Quad(lds_uri, ONTO_PRED, onto, lds_uri))

        anzo_client.add(qs)

    @classmethod
    def _register_lds_graph(cls, anzo_client: AnzoClient, lds_uri: str):
        qs = cls._lds_registry_statements(lds_uri)
        anzo_client.add(qs)

    @classmethod
    def _deregister_lds_graph(cls, anzo_client: AnzoClient, lds_uri: str):
        qs = cls._lds_registry_statements(lds_uri)
        anzo_client.remove(qs)

    @staticmethod
    def _lds_registry_statements(lds_uri: str) -> QuadStore:
        LDS_REG_URI = "http://cambridgesemantics.com/registries/LinkedDataSets"  # noqa
        DS_REG_URI = "http://cambridgesemantics.com/registries/DataSets"

        qs = QuadStore()
        qs.add(Quad(LDS_REG_URI, DNG_PRED, lds_uri, LDS_REG_URI))
        qs.add(Quad(DS_REG_URI, DNG_PRED, lds_uri, DS_REG_URI))
        return qs


def magic_triple(graph_uri: str) -> Quad:
    """Utility method to construct a Quad representing the "magic triple" for
    the given graph URI.

    Args:
        graph_uri: URI of the graph

    Returns:
        Quad representing the "magic triple" for the given graph URI
    """
    return Quad(graph_uri, TYPE_PRED, NG_TYPE,
                md_graph(graph_uri))


def md_graph(graph_uri: str) -> str:
    """Utility method to construct the metadata graph URI for the given graph URI.

    Args:
        graph_uri: URI of the graph

    Returns:
        The metadata graph URI for the given graph URI
    """
    encoded = urllib.parse.quote(graph_uri, safe="")
    return "http://openanzo.org/metadataGraphs(%s)" % encoded



# /*******************************************************************************
#  * Copyright (c) 2019 - 2022 Cambridge Semantics Incorporated.
#  * All rights reserved.
#  * 
#  * Contributors:
#  *     Cambridge Semantics Incorporated
#  *******************************************************************************/

import unittest
import time
import os

from pyanzo import (
    AnzoClient,
    QueryResult,
    QuadStore,
    Quad,
    TYPE_PRED,
)

from .test_common import (
    GRAPHMART,
    SERVER,
    PORT,
    USERNAME,
    PASSWORD
)


class TestAnzoClientQueryGraphmart(unittest.TestCase):
    # Paramterized values for the tests. This value is overwritten by the
    # values in the decorator

    graphmart = GRAPHMART

    # A simple select query used for testing core functionality in the tests
    select_query = """
        PREFIX data: <http://cambridgesemantics.com/ont/autogen/a69a/PyAnzo_Dictionary/PyAnzo_Datasource#>

        SELECT ?city WHERE {
            ?s a data:Person ;
                data:Person_City ?city .
        } ORDER BY ?city LIMIT 5
    """  # noqa

    select_query_results_first_layer = [
        ["Cambridge"],
        ["Somerville"],
        ["Watertown"],
    ]

    select_query_results_second_layer = [
        ["Boston"],
        ["Medford"],
    ]

    select_query_results = [
        ["Boston"],
        ["Cambridge"],
        ["Medford"],
        ["Somerville"],
        ["Watertown"],
    ]

    construct_query = """
        PREFIX data: <http://cambridgesemantics.com/ont/autogen/a69a/PyAnzo_Dictionary/PyAnzo_Datasource#>

        CONSTRUCT {
            ?s a data:Person .
        } WHERE {
            ?s a data:Person ;
                data:Person_City ?city .
        } ORDER BY ?city LIMIT 1
    """  # noqa

    construct_query_results = QuadStore({
        Quad(
            "http://csi.com/Person/879d629b-1015-45b9-8457-7271c0f2a9f0",
            TYPE_PRED,
            "http://cambridgesemantics.com/ont/autogen/a69a/PyAnzo_Dictionary/PyAnzo_Datasource#Person",  # noqa
            ""
        )
    })

    layer_one_uri = "http://cambridgesemantics.com/Layer/dd29f470d8d44dad8e6cf9ce3b6322fd"  # noqa
    layer_two_uri = "http://cambridgesemantics.com/Layer/d2bb3f418d7942d8a85fdedad608ccb3"  # noqa

    # TODO: WHY ISNT THIS USED ANYWHERE?
    # A slightly more complicated query,
    # useful for checking behavior of unbound variables
    select_query_with_optional = """
        PREFIX data: <http://cambridgesemantics.com/ont/autogen/a69a/PyAnzo_Dictionary/PyAnzo_Datasource#>

        SELECT ?city ?height WHERE {
          ?s a data:Person ;
               data:Person_Name ?name ;
               data:Person_City ?city ;
               data:Person_Age ?age .

          # person height is not a property and
          # will be unbound by design (for testing)
          OPTIONAL { ?s  data:Person_Height  ?height }

        } ORDER BY ?city LIMIT 5
     """  # noqa

    # TODO: is this not used anywhere??
    # The expected result of self.select_query_with_optional
    select_query_with_optional_results = QueryResult("""{
        "head": {"vars": ["city", "height"]},
        "results": {
            "bindings": [
                {"city": {"type": "literal", "value": "Boston"}},
                {"city": {"type": "literal", "value": "Cambridge"}},
                {"city": {"type": "literal", "value": "Medford"}},
                {"city": {"type": "literal", "value": "Somerville"}},
                {"city": {"type": "literal", "value": "Watertown"}}
            ]
        }
    }""")

    # Test with query files
    select_query_file = os.path.join(
        "tests", "test_assets", "simple_graphmart_query.rq"
    )

    malformed_query_file = os.path.join(
        "tests", "test_assets", "malformed_graphmart_query.rq"
    )

    # The test docstrings are convenient when looking at the code, but
    # they make the verbose testing view look cluttered. This function
    # removes comments/descriptions from the verbose test view.
    def shortDescription(self) -> str:
        return None

    def setUp(self) -> None:
        self.anzo_client = AnzoClient(
            server=SERVER, username=USERNAME, password=PASSWORD, port=PORT
        )

    @unittest.skip("Requires Anzo improvement to pass")
    def test_with_ask_query_true(self) -> None:
        query = "ASK WHERE { ?s ?p ?o }"
        result = self.anzo_client.query_graphmart(
            self.graphmart, query
        )

        self.assertEqual(
            result.as_table_results().as_list(), ["true"]
        )

    @unittest.skip("Requires Anzo improvement to pass")
    def test_with_ask_query_false(self) -> None:
        query = "ASK WHERE { <urn://doesnt/exist> ?p ?o }"
        result = self.anzo_client.query_graphmart(
            self.graphmart, query
        )

        self.assertEqual(
            result.as_table_results().as_list(), ["false"]
        )

    def test_malformed_graphmart_query_with_file(self) -> None:
        self.assertRaises(
            RuntimeError, self.anzo_client.query_graphmart, self.graphmart,
            query_file=self.malformed_query_file
        )

    def test_bad_graphmart_query_with_file(self) -> None:
        self.assertRaises(
            RuntimeError, self.anzo_client.query_graphmart,
            self.graphmart, query_file="abcqwerty.qwerty"
        )

    def test_malformed_graphmart_query(self) -> None:
        bad_query = "SE"
        self.assertRaises(
            RuntimeError, self.anzo_client.query_graphmart,
            self.graphmart, query_string=bad_query
        )

    def test_empty_graphmart_query_string(self) -> None:
        self.assertRaises(
            ValueError, self.anzo_client.query_graphmart,
            self.graphmart, query_string=""
        )

    def test_empty_graphmart(self) -> None:
        self.assertRaises(
            ValueError, self.anzo_client.query_graphmart,
            "", query_string=self.select_query
        )

    def test_graphmart_with_no_query(self) -> None:
        self.assertRaises(
            ValueError, self.anzo_client.query_graphmart, self.graphmart
        )

    def test_bad_graphmart_in_graphmart_query(self) -> None:
        bad_graphmart = "qwerty"
        self.assertRaises(
            RuntimeError, self.anzo_client.query_graphmart,
            bad_graphmart, self.select_query_file
        )

    def test_basic_query(self) -> None:
        result = self.anzo_client.query_graphmart(
            self.graphmart,
            query_string=self.select_query
        )

        self.assertEqual(
            result.as_table_results().as_list(), self.select_query_results
        )

    def test_basic_construct_query(self) -> None:
        result = self.anzo_client.query_graphmart(
            self.graphmart, self.construct_query
        )

        self.assertEqual(
            result.as_quad_store(), self.construct_query_results
        )

    def test_construct_query_with_no_results(self) -> None:
        query = "CONSTRUCT { ?s a <urn://type> } WHERE { ?s a <urn://not/a/type> } LIMIT 1"  # noqa
        result = self.anzo_client.query_graphmart(
            self.graphmart, query
        )

        self.assertEqual(
            result.as_quad_store(), QuadStore()
        )

    def test_with_first_layer_specified(self) -> None:
        result = self.anzo_client.query_graphmart(
            self.graphmart,
            query_string=self.select_query,
            data_layers=[self.layer_one_uri]
        )

        self.assertEqual(
            result.as_table_results().as_list(),
            self.select_query_results_first_layer
        )

    def test_with_first_second_specified(self) -> None:
        result = self.anzo_client.query_graphmart(
            self.graphmart,
            query_string=self.select_query,
            data_layers=[self.layer_two_uri]
        )

        self.assertEqual(
            result.as_table_results().as_list(),
            self.select_query_results_second_layer
        )

    def test_with_both_layers_specified(self) -> None:
        result = self.anzo_client.query_graphmart(
            self.graphmart,
            query_string=self.select_query,
            data_layers=[self.layer_one_uri, self.layer_two_uri]
        )

        self.assertEqual(
            result.as_table_results().as_list(), self.select_query_results
        )

    def test_with_cache_skipped(self) -> None:
        try:
            self.anzo_client.query_graphmart(self.graphmart,
                                             query_string=self.select_query,
                                             skip_cache=True)
        except RuntimeError:
            self.fail("Cache Skipping Not Implemented")

    def test_timeout(self) -> None:
        ac = AnzoClient(
            server=SERVER, username=USERNAME, password=PASSWORD, port=PORT,
            timeout_seconds=0.00001
        )

        self.assertRaises(
            TimeoutError, ac.query_graphmart,
            self.graphmart, query_string=self.select_query
        )
